@@ -1,21 +1,48 @@
 //! Types for working with [Anim] data in .nuanmb files.
-use binread::{io::StreamPosition, BinRead};
+//!
+//! # Examples
+//! Animation data is stored in a heirarchy.
+//! Values for each frame are stored at the [TrackData] level.
+/*!
+```rust no_run
+# fn main() -> Result<(), Box<dyn std::error::Error>> {
+use ssbh_data::prelude::*;
+
+let anim = AnimData::from_file("model.nuanmb")?;
+
+for group in anim.groups {
+    for node in group.nodes {
+        for track in node.tracks {
+            println!("Frame Count: {}", track.values.len());
+        }
+    }
+}
+# Ok(()) }
+```
+ */
+//! # File Differences
+//! Unmodified files are not guaranteed to be binary identical after saving.
+//! Compressed animations use lossy compression for all data types except [TrackValues::Boolean].
+//! When converting to [Anim], compression is enabled for a track if compression would save space.
+//! This may produce differences with the original due to compression differences.
+//! These errors are small in practice but may cause gameplay differences such as online desyncs.
+use binread::{io::StreamPosition, BinRead, BinReaderExt};
 use std::{
     convert::{TryFrom, TryInto},
     error::Error,
-    io::{Cursor, Read, Seek, Write},
-    path::Path,
+    io::{Cursor, Write},
 };
 
 use ssbh_write::SsbhWrite;
 
-use ssbh_lib::formats::anim::{
-    Anim, AnimGroup, AnimHeader, AnimHeaderV20, AnimHeaderV21, AnimNode, AnimTrackV2,
-    CompressionType, TrackFlags, TrackType, UnkData, UnkTrackFlags,
+use ssbh_lib::{
+    formats::anim::{
+        Anim, CompressionType, Group, Node, TrackFlags, TrackTypeV2, TrackV2,
+        TransformFlags as AnimTransformFlags, UnkData,
+    },
+    Version,
 };
 
-use thiserror::Error;
-
 #[cfg(feature = "serde")]
 use serde::{Deserialize, Serialize};
 
@@ -23,51 +50,29 @@ pub use ssbh_lib::{formats::anim::GroupType, Vector3, Vector4};
 
 mod buffers;
 use buffers::*;
+mod bitutils;
 mod compression;
 
-use crate::SsbhData;
-
-// TODO: Add module level documentation to show anim <-> data conversions and describe overall structure and design.
-
-/// The data associated with an [Anim] file.
+/// Data associated with an [Anim] file.
 /// Supported versions are 2.0 and 2.1.
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct AnimData {
     pub major_version: u16,
     pub minor_version: u16,
-    // TODO: Is float the best choice here?
-    // TODO: Just use a usize but return an error on invalid lengths?
+
     /// The index of the last frame in the animation,
     /// which is calculated as `(frame_count - 1) as f32`.
     ///
     /// Constant animations will last for final_frame_index + 1 many frames.
+    ///
+    /// Frames use floating point to allow the rendering speed to differ from the animation speed.
+    /// For example, some animations in Smash Ultimate interpolate when playing the game at 60fps but 1/4 speed.
     pub final_frame_index: f32,
     pub groups: Vec<GroupData>,
 }
 
-impl SsbhData for AnimData {
-    type WriteError = AnimError;
-
-    fn from_file<P: AsRef<Path>>(path: P) -> Result<Self, Box<dyn std::error::Error>> {
-        Anim::from_file(path)?.try_into()
-    }
-
-    fn read<R: Read + Seek>(reader: &mut R) -> Result<Self, Box<dyn std::error::Error>> {
-        Anim::read(reader)?.try_into()
-    }
-
-    fn write<W: std::io::Write + Seek>(&self, writer: &mut W) -> Result<(), AnimError> {
-        Anim::try_from(self)?.write(writer)?;
-        Ok(())
-    }
-
-    fn write_to_file<P: AsRef<Path>>(&self, path: P) -> Result<(), AnimError> {
-        Anim::try_from(self)?.write_to_file(path)?;
-        Ok(())
-    }
-}
-
 // TODO: Test these conversions.
 impl TryFrom<Anim> for AnimData {
     type Error = Box<dyn Error>;
@@ -81,72 +86,125 @@ impl TryFrom<&Anim> for AnimData {
     type Error = Box<dyn Error>;
 
     fn try_from(anim: &Anim) -> Result<Self, Self::Error> {
+        let (major_version, minor_version) = anim.major_minor_version();
         Ok(Self {
-            major_version: anim.major_version,
-            minor_version: anim.minor_version,
-            final_frame_index: match &anim.header {
-                AnimHeader::HeaderV1(h) => h.final_frame_index,
-                AnimHeader::HeaderV20(h) => h.final_frame_index,
-                AnimHeader::HeaderV21(h) => h.final_frame_index,
+            major_version,
+            minor_version,
+            final_frame_index: match &anim {
+                Anim::V12 {
+                    final_frame_index, ..
+                } => *final_frame_index,
+                Anim::V20 {
+                    final_frame_index, ..
+                } => *final_frame_index,
+                Anim::V21 {
+                    final_frame_index, ..
+                } => *final_frame_index,
             },
             groups: read_anim_groups(anim)?,
         })
     }
 }
 
+impl TryFrom<AnimData> for Anim {
+    type Error = error::Error;
+
+    fn try_from(data: AnimData) -> Result<Self, Self::Error> {
+        create_anim(&data)
+    }
+}
+
 impl TryFrom<&AnimData> for Anim {
-    type Error = AnimError;
+    type Error = error::Error;
 
     fn try_from(data: &AnimData) -> Result<Self, Self::Error> {
         create_anim(data)
     }
 }
 
-/// Errors while creating an [Anim] from [AnimData].
-#[derive(Error, Debug)]
+pub mod error {
+    use super::*;
+    use thiserror::Error;
 
-pub enum AnimError {
-    /// Creating an [Anim] file for the given version is not supported.
-    #[error(
-        "Creating a version {}.{} anim is not supported.",
-        major_version,
-        minor_version
-    )]
-    UnsupportedVersion {
-        major_version: u16,
-        minor_version: u16,
-    },
+    /// Errors while creating an [Anim] from [AnimData].
+    #[derive(Debug, Error)]
+    pub enum Error {
+        /// Creating an [Anim] file for the given version is not supported.
+        #[error(
+            "Creating a version {}.{} anim is not supported.",
+            major_version,
+            minor_version
+        )]
+        UnsupportedVersion {
+            major_version: u16,
+            minor_version: u16,
+        },
 
-    /// The final frame index is negative or smaller than the
-    // index of the final frame in the longest track.
-    #[error(
-        "Final frame index {} must be non negative and at least as 
-         large as the index of the final frame in the longest track.",
-        final_frame_index
-    )]
-    InvalidFinalFrameIndex { final_frame_index: f32 },
+        /// The final frame index is negative or smaller than the
+        // index of the final frame in the longest track.
+        #[error(
+            "Final frame index {} must be non negative and at least as 
+             large as the index of the final frame in the longest track.",
+            final_frame_index
+        )]
+        InvalidFinalFrameIndex { final_frame_index: f32 },
 
-    #[error(
-        "Scale options of {:?} cannot be preserved for a {} track.",
-        scale_options,
-        if *compressed {"compressed"} else { "uncompressed"}
-    )]
-    UnsupportedTrackScaleOptions {
-        scale_options: ScaleOptions,
-        compressed: bool,
-    },
+        #[error(
+            "Scale options of {:?} cannot be preserved for a {} track.",
+            scale_options,
+            if *compressed {"compressed"} else { "uncompressed"}
+        )]
+        UnsupportedTrackScaleOptions {
+            scale_options: ScaleOptions,
+            compressed: bool,
+        },
 
-    /// An error occurred while writing data to a buffer.
-    #[error(transparent)]
-    Io(#[from] std::io::Error),
+        /// An error occurred while writing data to a buffer.
+        #[error(transparent)]
+        Io(#[from] std::io::Error),
 
-    /// An error occurred while reading data from a buffer.
-    #[error(transparent)]
-    BinRead(#[from] binread::error::Error),
+        /// An error occurred while reading data from a buffer.
+        #[error(transparent)]
+        BinRead(#[from] binread::error::Error),
 
-    /// An error occurred while reading compressed data from a buffer.
-    #[error(transparent)]
-    BitError(#[from] bitbuffer::BitError),
+        /// An error occurred while reading compressed data from a buffer.
+        #[error(transparent)]
+        BitError(#[from] bitutils::BitReadError),
+
+        #[error(
+            "Compressed header bits per entry of {} does not match expected value of {}.",
+            actual,
+            expected
+        )]
+        UnexpectedBitCount { expected: usize, actual: usize },
+
+        #[error(
+            "Track data range {0}..{0}+{1} is out of range for a buffer of size {2}.",
+            start,
+            size,
+            buffer_size
+        )]
+        InvalidTrackDataRange {
+            start: usize,
+            size: usize,
+            buffer_size: usize,
+        },
+
+        /// The buffer index is not valid for a version 1.2 anim file.
+        #[error(
+            "Buffer index {} is out of range for a buffer collection of size {}.",
+            buffer_index,
+            buffer_count
+        )]
+        BufferIndexOutOfRange {
+            buffer_index: usize,
+            buffer_count: usize,
+        },
+
+        /// An error occurred while reading the compressed header for version 2.0 or later.
+        #[error("The track data compression header is malformed and cannot be read.")]
+        MalformedCompressionHeader,
+    }
 }
 
 enum AnimVersion {
@@ -155,11 +213,11 @@ enum AnimVersion {
 }
 
 // TODO: Test this for a small example?
-fn create_anim(data: &AnimData) -> Result<Anim, AnimError> {
+fn create_anim(data: &AnimData) -> Result<Anim, error::Error> {
     let version = match (data.major_version, data.minor_version) {
         (2, 0) => Ok(AnimVersion::Version20),
         (2, 1) => Ok(AnimVersion::Version21),
-        _ => Err(AnimError::UnsupportedVersion {
+        _ => Err(error::Error::UnsupportedVersion {
             major_version: data.major_version,
             minor_version: data.minor_version,
         }),
@@ -191,13 +249,13 @@ fn create_anim(data: &AnimData) -> Result<Anim, AnimError> {
     {
         Ok(data.final_frame_index)
     } else {
-        Err(AnimError::InvalidFinalFrameIndex {
+        Err(error::Error::InvalidFinalFrameIndex {
             final_frame_index: data.final_frame_index,
         })
     }?;
 
-    let header = match version {
-        AnimVersion::Version20 => AnimHeader::HeaderV20(AnimHeaderV20 {
+    match version {
+        AnimVersion::Version20 => Ok(Anim::V20 {
             final_frame_index,
             unk1: 1,
             unk2: 3,
@@ -205,7 +263,7 @@ fn create_anim(data: &AnimData) -> Result<Anim, AnimError> {
             groups: animations.into(),
             buffer: buffer.into_inner().into(),
         }),
-        AnimVersion::Version21 => AnimHeader::HeaderV21(AnimHeaderV21 {
+        AnimVersion::Version21 => Ok(Anim::V21 {
             final_frame_index,
             unk1: 1,
             unk2: 3,
@@ -218,19 +276,11 @@ fn create_anim(data: &AnimData) -> Result<Anim, AnimError> {
                 unk2: Vec::new().into(),
             },
         }),
-    };
-
-    // TODO: Check that the header matches the version number?
-    let anim = Anim {
-        major_version: data.major_version,
-        minor_version: data.minor_version,
-        header,
-    };
-    Ok(anim)
+    }
 }
 
-fn create_anim_group(g: &GroupData, buffer: &mut Cursor<Vec<u8>>) -> Result<AnimGroup, AnimError> {
-    Ok(AnimGroup {
+fn create_anim_group(g: &GroupData, buffer: &mut Cursor<Vec<u8>>) -> Result<Group, error::Error> {
+    Ok(Group {
         group_type: g.group_type,
         nodes: g
             .nodes
@@ -241,8 +291,8 @@ fn create_anim_group(g: &GroupData, buffer: &mut Cursor<Vec<u8>>) -> Result<Anim
     })
 }
 
-fn create_anim_node(n: &NodeData, buffer: &mut Cursor<Vec<u8>>) -> Result<AnimNode, AnimError> {
-    Ok(AnimNode {
+fn create_anim_node(n: &NodeData, buffer: &mut Cursor<Vec<u8>>) -> Result<Node, error::Error> {
+    Ok(Node {
         name: n.name.as_str().into(), // TODO: Make a convenience method for this?
         tracks: n
             .tracks
@@ -256,7 +306,7 @@ fn create_anim_node(n: &NodeData, buffer: &mut Cursor<Vec<u8>>) -> Result<AnimNo
 fn create_anim_track_v2(
     buffer: &mut Cursor<Vec<u8>>,
     t: &TrackData,
-) -> Result<AnimTrackV2, AnimError> {
+) -> Result<TrackV2, error::Error> {
     let compression_type = infer_optimal_compression_type(&t.values);
 
     // The current stream position matches the offsets used for Smash Ultimate's anim files.
@@ -278,14 +328,14 @@ fn create_anim_track_v2(
     buffer.write_all(&track_data.into_inner())?;
     let pos_after = buffer.stream_pos()?;
 
-    Ok(AnimTrackV2 {
+    Ok(TrackV2 {
         name: t.name.as_str().into(),
         flags: TrackFlags {
             track_type: t.values.track_type(),
             compression_type,
         },
         frame_count: t.values.len() as u32,
-        unk_flags: UnkTrackFlags::new(), // TODO: preserve these flags?
+        transform_flags: t.transform_flags.into(),
         data_offset: pos_before as u32,
         data_size: pos_after - pos_before,
     })
@@ -316,29 +366,83 @@ fn infer_optimal_compression_type(values: &TrackValues) -> CompressionType {
 }
 
 // TODO: Test conversions from anim?
-fn read_anim_groups(anim: &Anim) -> Result<Vec<GroupData>, AnimError> {
-    match &anim.header {
+fn read_anim_groups(anim: &Anim) -> Result<Vec<GroupData>, error::Error> {
+    match anim {
         // TODO: Create fake groups for version 1.0?
-        ssbh_lib::formats::anim::AnimHeader::HeaderV1(_) => Err(AnimError::UnsupportedVersion {
-            major_version: anim.major_version,
-            minor_version: anim.minor_version,
-        }),
-        ssbh_lib::formats::anim::AnimHeader::HeaderV20(header) => {
-            read_anim_groups_v20(&header.groups.elements, &header.buffer.elements)
+        ssbh_lib::prelude::Anim::V12 {
+            tracks, buffers, ..
+        } => {
+            // TODO: Group by type?
+            // TODO: Assign a single node to each track with the track name as the name?
+            // TODO: Use the track type as the track name like "Transform"?
+            for track in &tracks.elements {
+                create_track_data_v12(track, buffers)?;
+            }
+            Ok(Vec::new())
         }
-        ssbh_lib::formats::anim::AnimHeader::HeaderV21(header) => {
-            read_anim_groups_v20(&header.groups.elements, &header.buffer.elements)
+        ssbh_lib::formats::anim::Anim::V20 { groups, buffer, .. } => {
+            read_groups_v20(&groups.elements, &buffer.elements)
+        }
+        ssbh_lib::formats::anim::Anim::V21 { groups, buffer, .. } => {
+            read_groups_v20(&groups.elements, &buffer.elements)
         }
     }
 }
 
-fn read_anim_groups_v20(
-    anim_groups: &[ssbh_lib::formats::anim::AnimGroup],
+fn create_track_data_v12(
+    track: &ssbh_lib::formats::anim::TrackV1,
+    buffers: &ssbh_lib::SsbhArray<ssbh_lib::SsbhByteBuffer>,
+) -> Result<TrackData, error::Error> {
+    // TODO: Add tests for this to buffers.rs.
+    for property in &track.properties.elements {
+        let data = buffers.elements.get(property.buffer_index as usize).ok_or(
+            error::Error::BufferIndexOutOfRange {
+                buffer_index: property.buffer_index as usize,
+                buffer_count: buffers.elements.len(),
+            },
+        )?;
+
+        let mut reader = Cursor::new(&data.elements);
+        let header: u32 = reader.read_le()?;
+
+        println!("{:?},{:x?}", property.name.to_string_lossy(), header);
+
+        match header {
+            0x1003 => {
+                println!("{:x?}", reader.read_le::<f32>()?);
+            }
+            0x2003 => {
+                println!("{:?}", reader.read_le::<(f32, f32)>()?);
+            }
+            0x3003 => {
+                println!("{:?}", reader.read_le::<Vector3>()?);
+            }
+            0x4003 => {
+                println!("{:?}", reader.read_le::<Vector4>()?);
+            }
+            0x1013 => {
+                println!("{:x?}", reader.read_le::<u16>()?);
+            }
+            x => println!("Unrecognized header: {:?}", x),
+        }
+    }
+    println!();
+
+    // TODO: Set the track data based on type?
+    Ok(TrackData {
+        name: track.name.to_string_lossy(),
+        scale_options: ScaleOptions::default(),
+        values: TrackValues::Float(Vec::new()),
+        transform_flags: TransformFlags::default(),
+    })
+}
+
+fn read_groups_v20(
+    anim_groups: &[ssbh_lib::formats::anim::Group],
     anim_buffer: &[u8],
-) -> Result<Vec<GroupData>, AnimError> {
+) -> Result<Vec<GroupData>, error::Error> {
     let mut groups = Vec::new();
 
-    // TODO: Return a more meaningful error type.
     for anim_group in anim_groups {
         let mut nodes = Vec::new();
 
@@ -367,64 +471,92 @@ fn read_anim_groups_v20(
     Ok(groups)
 }
 
-// TODO: Add tests for preserving scale inheritance and compensate scale.
 fn create_track_data_v20(
-    anim_track: &ssbh_lib::formats::anim::AnimTrackV2,
-    anim_buffer: &[u8],
-) -> Result<TrackData, AnimError> {
-    let start = anim_track.data_offset as usize;
-    let end = start + anim_track.data_size as usize;
-    let buffer = &anim_buffer[start..end];
+    track: &ssbh_lib::formats::anim::TrackV2,
+    buffer: &[u8],
+) -> Result<TrackData, error::Error> {
+    let start = track.data_offset as usize;
+    let end =
+        start
+            .checked_add(track.data_size as usize)
+            .ok_or(error::Error::InvalidTrackDataRange {
+                start: track.data_offset as usize,
+                size: track.data_size as usize,
+                buffer_size: buffer.len(),
+            })?;
+    let buffer = buffer
+        .get(start..end)
+        .ok_or(error::Error::InvalidTrackDataRange {
+            start: track.data_offset as usize,
+            size: track.data_size as usize,
+            buffer_size: buffer.len(),
+        })?;
+
     let (values, inherit_scale, compensate_scale) =
-        read_track_values(buffer, anim_track.flags, anim_track.frame_count as usize)?;
+        read_track_values(buffer, track.flags, track.frame_count as usize)?;
+
+    // The compensate scale override is included in scale options instead.
     Ok(TrackData {
-        name: anim_track.name.to_string_lossy(),
+        name: track.name.to_string_lossy(),
         values,
         scale_options: ScaleOptions {
             inherit_scale,
-            compensate_scale,
+            compensate_scale: compensate_scale
+                && !track.transform_flags.override_compensate_scale(),
         },
+        transform_flags: track.transform_flags.into(),
     })
 }
 
+/// Data associated with a [Group].
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct GroupData {
     /// The usage type for all the [NodeData] in [nodes](#structfield.nodes)
     pub group_type: GroupType,
     pub nodes: Vec<NodeData>,
 }
 
+/// Data associated with a [Node].
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct NodeData {
     pub name: String,
     pub tracks: Vec<TrackData>,
 }
 
-/// The data associated with an [AnimTrackV2].
+/// The data associated with a [TrackV2].
 ///
 /// # Examples
-/// The scale settings can usually be left at the default value.
+/// The scale settings and transform flags should usually use their default value.
 /**
 ```rust
-use ssbh_data::anim_data::{TrackData, TrackValues, ScaleOptions, Transform};
+use ssbh_data::anim_data::{TrackData, TrackValues, ScaleOptions, Transform, TransformFlags};
 
 let track = TrackData {
     name: "Transform".to_string(),
     values: TrackValues::Transform(vec![Transform::IDENTITY]),
-    scale_options: ScaleOptions::default()
+    scale_options: ScaleOptions::default(),
+    transform_flags: TransformFlags::default()
 };
 ```
  */
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct TrackData {
     /// The name of the property to animate.
+    ///
+    /// For tracks in a group of type [GroupType::Material], this is the name of the material parameter like "CustomVector31".
+    /// Other group types tend to use the name of the group type like "Transform" or "Visibility".
     pub name: String,
 
     pub scale_options: ScaleOptions,
 
+    pub transform_flags: TransformFlags,
+
     /// The frame values for the property specified by [name](#structfield.name).
     ///
     /// Each element in the [TrackValues] provides the value for a single frame.
@@ -434,12 +566,14 @@ pub struct TrackData {
     pub values: TrackValues,
 }
 
+/// Determines how scaling is calculated for bone chains. Only applies to [TrackValues::Transform].
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Eq, Clone, Copy)]
 pub struct ScaleOptions {
     /// Accumulate the parent's scaling when `true`.
     ///
-    /// The global scale in world space, is `current_scale * parent_scale` applied recursively on the parent.
+    /// The global scale in world space is `current_scale * parent_scale` applied recursively on the parent.
     pub inherit_scale: bool,
 
     /// Revert the scaling of the immediate parent when `true`.
@@ -461,8 +595,40 @@ impl Default for ScaleOptions {
     }
 }
 
-// TODO: Investigate if the names based on the Anim 1.2 property names are accurate.
+/// See [ssbh_lib::formats::anim::TransformFlags].
+// Including compensate scale would be redundant with ScaleOptions.
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Eq, Default, Clone, Copy)]
+pub struct TransformFlags {
+    pub override_translation: bool,
+    pub override_rotation: bool,
+    pub override_scale: bool,
+}
+
+impl From<TransformFlags> for AnimTransformFlags {
+    fn from(f: TransformFlags) -> Self {
+        Self::new()
+            .with_override_translation(f.override_translation)
+            .with_override_rotation(f.override_rotation)
+            .with_override_scale(f.override_scale)
+    }
+}
+
+impl From<AnimTransformFlags> for TransformFlags {
+    fn from(f: AnimTransformFlags) -> Self {
+        Self {
+            override_translation: f.override_translation(),
+            override_rotation: f.override_rotation(),
+            override_scale: f.override_compensate_scale(),
+        }
+    }
+}
+
+// TODO: Investigate if the names based on the Anim 1.2 property names are accurate.
+/// A decomposed 2D transformation for texture coordinates.
+#[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[derive(Debug, BinRead, PartialEq, SsbhWrite, Default, Clone, Copy)]
 pub struct UvTransform {
     pub scale_u: f32,
@@ -472,8 +638,9 @@ pub struct UvTransform {
     pub translate_v: f32,
 }
 
-/// A decomposed transformation consisting of a scale, rotation, and translation.
+/// A decomposed 3D transformation consisting of a scale, rotation, and translation.
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[derive(Debug, PartialEq, Clone, Copy, Default)]
 pub struct Transform {
     /// XYZ scale
@@ -509,7 +676,8 @@ impl Transform {
 
 /// A value collection with an element for each frame of the animation.
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug, PartialEq)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub enum TrackValues {
     /// Transformations used for camera or skeletal animations.
     Transform(Vec<Transform>),
@@ -562,14 +730,14 @@ impl TrackValues {
         }
     }
 
-    fn track_type(&self) -> TrackType {
+    fn track_type(&self) -> TrackTypeV2 {
         match self {
-            TrackValues::Transform(_) => TrackType::Transform,
-            TrackValues::UvTransform(_) => TrackType::UvTransform,
-            TrackValues::Float(_) => TrackType::Float,
-            TrackValues::PatternIndex(_) => TrackType::PatternIndex,
-            TrackValues::Boolean(_) => TrackType::Boolean,
-            TrackValues::Vector4(_) => TrackType::Vector4,
+            TrackValues::Transform(_) => TrackTypeV2::Transform,
+            TrackValues::UvTransform(_) => TrackTypeV2::UvTransform,
+            TrackValues::Float(_) => TrackTypeV2::Float,
+            TrackValues::PatternIndex(_) => TrackTypeV2::PatternIndex,
+            TrackValues::Boolean(_) => TrackTypeV2::Boolean,
+            TrackValues::Vector4(_) => TrackTypeV2::Vector4,
         }
     }
 }
@@ -578,7 +746,6 @@ impl TrackValues {
 mod tests {
     use crate::assert_hex_eq;
     use hexlit::hex;
-    use ssbh_lib::formats::anim::AnimHeaderV21;
 
     use super::*;
 
@@ -595,11 +762,11 @@ mod tests {
         .unwrap();
 
         assert!(matches!(
-            anim.header,
-            AnimHeader::HeaderV20(AnimHeaderV20 {
+            anim,
+            Anim::V20 {
                 final_frame_index,
                 ..
-            }) if final_frame_index == 1.5
+            } if final_frame_index == 1.5
         ));
     }
 
@@ -613,10 +780,10 @@ mod tests {
         })
         .unwrap();
 
-        assert!(matches!(anim.header, AnimHeader::HeaderV21(AnimHeaderV21 {
+        assert!(matches!(anim, Anim::V21 {
             final_frame_index, 
             ..
-        }) if final_frame_index == 2.5));
+        } if final_frame_index == 2.5));
     }
 
     #[test]
@@ -630,7 +797,7 @@ mod tests {
 
         assert!(matches!(
             result,
-            Err(AnimError::InvalidFinalFrameIndex {
+            Err(error::Error::InvalidFinalFrameIndex {
                 final_frame_index
             }) if final_frame_index == -1.0
         ));
@@ -650,6 +817,7 @@ mod tests {
                         name: String::new(),
                         values: TrackValues::Boolean(vec![true; 4]),
                         scale_options: ScaleOptions::default(),
+                        transform_flags: TransformFlags::default(),
                     }],
                 }],
             }],
@@ -658,7 +826,7 @@ mod tests {
         // A value of at least 3.0 is expected.
         assert!(matches!(
             result,
-            Err(AnimError::InvalidFinalFrameIndex {
+            Err(error::Error::InvalidFinalFrameIndex {
                 final_frame_index
             }) if final_frame_index == 2.0
         ));
@@ -674,10 +842,10 @@ mod tests {
         })
         .unwrap();
 
-        assert!(matches!(anim.header, AnimHeader::HeaderV21(AnimHeaderV21 {
+        assert!(matches!(anim, Anim::V21 {
             final_frame_index, 
             ..
-        }) if final_frame_index == 0.0));
+        } if final_frame_index == 0.0));
     }
 
     #[test]
@@ -691,7 +859,7 @@ mod tests {
 
         assert!(matches!(
             result,
-            Err(AnimError::UnsupportedVersion {
+            Err(error::Error::UnsupportedVersion {
                 major_version: 1,
                 minor_version: 2
             })
@@ -721,11 +889,13 @@ mod tests {
                     name: "t1".to_string(),
                     values: TrackValues::Float(vec![1.0, 2.0, 3.0]),
                     scale_options: ScaleOptions::default(),
+                    transform_flags: TransformFlags::default(),
                 },
                 TrackData {
                     name: "t2".to_string(),
                     values: TrackValues::PatternIndex(vec![4, 5]),
                     scale_options: ScaleOptions::default(),
+                    transform_flags: TransformFlags::default(),
                 },
             ],
         };
@@ -740,7 +910,7 @@ mod tests {
         assert_eq!("t1", t1.name.to_str().unwrap());
         assert_eq!(
             TrackFlags {
-                track_type: TrackType::Float,
+                track_type: TrackTypeV2::Float,
                 compression_type: CompressionType::Direct
             },
             t1.flags
@@ -753,7 +923,7 @@ mod tests {
         assert_eq!("t2", t2.name.to_str().unwrap());
         assert_eq!(
             TrackFlags {
-                track_type: TrackType::PatternIndex,
+                track_type: TrackTypeV2::PatternIndex,
                 compression_type: CompressionType::Direct
             },
             t2.flags
@@ -944,6 +1114,87 @@ mod tests {
         );
     }
 
+    #[test]
+    fn read_v20_track_invalid_offset() {
+        let result = create_track_data_v20(
+            &TrackV2 {
+                name: "abc".into(),
+                flags: TrackFlags {
+                    track_type: TrackTypeV2::Transform,
+                    compression_type: CompressionType::Compressed,
+                },
+                frame_count: 2,
+                transform_flags: AnimTransformFlags::new(),
+                data_offset: 5,
+                data_size: 1,
+            },
+            &[0u8; 4],
+        );
+
+        assert!(matches!(
+            result,
+            Err(error::Error::InvalidTrackDataRange {
+                start: 5,
+                size: 1,
+                buffer_size: 4
+            })
+        ));
+    }
+
+    #[test]
+    fn read_v20_track_offset_overflow() {
+        let result = create_track_data_v20(
+            &TrackV2 {
+                name: "abc".into(),
+                flags: TrackFlags {
+                    track_type: TrackTypeV2::Transform,
+                    compression_type: CompressionType::Compressed,
+                },
+                frame_count: 2,
+                transform_flags: AnimTransformFlags::new(),
+                data_offset: u32::MAX,
+                data_size: 1,
+            },
+            &[0u8; 4],
+        );
+
+        assert!(matches!(
+            result,
+            Err(error::Error::InvalidTrackDataRange {
+                start: 4294967295,
+                size: 1,
+                buffer_size: 4
+            })
+        ));
+    }
+
+    #[test]
+    fn read_v20_track_invalid_size() {
+        let result = create_track_data_v20(
+            &TrackV2 {
+                name: "abc".into(),
+                flags: TrackFlags {
+                    track_type: TrackTypeV2::Transform,
+                    compression_type: CompressionType::Compressed,
+                },
+                frame_count: 2,
+                transform_flags: AnimTransformFlags::new(),
+                data_offset: 0,
+                data_size: 5,
+            },
+            &[0u8; 3],
+        );
+
+        assert!(matches!(
+            result,
+            Err(error::Error::InvalidTrackDataRange {
+                start: 0,
+                size: 5,
+                buffer_size: 3
+            })
+        ));
+    }
+
     #[test]
     fn read_v20_track_compressed_inherit_scale() {
         // assist/shovelknight/model/body/c00/model.nuanmb, ArmL, Transform
@@ -972,14 +1223,14 @@ mod tests {
         );
 
         let data = create_track_data_v20(
-            &AnimTrackV2 {
+            &TrackV2 {
                 name: "abc".into(),
                 flags: TrackFlags {
-                    track_type: TrackType::Transform,
+                    track_type: TrackTypeV2::Transform,
                     compression_type: CompressionType::Compressed,
                 },
                 frame_count: 2,
-                unk_flags: UnkTrackFlags::new(),
+                transform_flags: AnimTransformFlags::new(),
                 data_offset: 0,
                 data_size: buffer.len() as u64,
             },
@@ -1030,6 +1281,7 @@ mod tests {
                     inherit_scale: true,
                     compensate_scale: false,
                 },
+                transform_flags: TransformFlags::default(),
             },
         )
         .unwrap();
@@ -1069,14 +1321,14 @@ mod tests {
         );
 
         let data = create_track_data_v20(
-            &AnimTrackV2 {
+            &TrackV2 {
                 name: "abc".into(),
                 flags: TrackFlags {
-                    track_type: TrackType::Transform,
+                    track_type: TrackTypeV2::Transform,
                     compression_type: CompressionType::Compressed,
                 },
                 frame_count: 2,
-                unk_flags: UnkTrackFlags::new(),
+                transform_flags: AnimTransformFlags::new(),
                 data_offset: 0,
                 data_size: buffer.len() as u64,
             },
@@ -1127,6 +1379,7 @@ mod tests {
                     inherit_scale: false,
                     compensate_scale: false,
                 },
+                transform_flags: TransformFlags::default(),
             },
         )
         .unwrap();
@@ -1139,6 +1392,38 @@ mod tests {
 
     #[test]
     fn read_v20_track_uncompressed() {
+        // assist/shovelknight/model/body/c00/model.nuanmb, FingerL11, Transform
+        let buffer = hex!(
+            0000803f 0000803f 0000803f          // scale
+            00000000 00000000 00000000          // translation
+            0000803f bea4c13f_79906ebe f641bebe // rotation
+            00000000                            // compensate scale
+        );
+
+        let data = create_track_data_v20(
+            &TrackV2 {
+                name: "abc".into(),
+                flags: TrackFlags {
+                    track_type: TrackTypeV2::Transform,
+                    compression_type: CompressionType::ConstTransform,
+                },
+                frame_count: 1,
+                transform_flags: AnimTransformFlags::new(),
+                data_offset: 0,
+                data_size: buffer.len() as u64,
+            },
+            &buffer,
+        )
+        .unwrap();
+
+        // TODO: This should test the values, but this overlaps with anim_buffer tests?
+        assert_eq!("abc", data.name);
+        assert_eq!(true, data.scale_options.inherit_scale);
+        assert_eq!(false, data.scale_options.compensate_scale);
+    }
+
+    #[test]
+    fn read_v20_track_uncompressed_compensate_scale() {
         // assist/shovelknight/model/body/c00/model.nuanmb, FingerL11, Transform
         let buffer = hex!(
             0000803f 0000803f 0000803f          // scale
@@ -1148,14 +1433,14 @@ mod tests {
         );
 
         let data = create_track_data_v20(
-            &AnimTrackV2 {
+            &TrackV2 {
                 name: "abc".into(),
                 flags: TrackFlags {
-                    track_type: TrackType::Transform,
+                    track_type: TrackTypeV2::Transform,
                     compression_type: CompressionType::ConstTransform,
                 },
                 frame_count: 1,
-                unk_flags: UnkTrackFlags::new(),
+                transform_flags: AnimTransformFlags::new(),
                 data_offset: 0,
                 data_size: buffer.len() as u64,
             },
@@ -1168,6 +1453,40 @@ mod tests {
         // Uncompressed transforms seem to inherit scale with ConstTransform.
         // TODO: Investigate if uncompressed transforms always inherit scale.
         assert_eq!(true, data.scale_options.inherit_scale);
+        assert_eq!(true, data.scale_options.compensate_scale);
+    }
+
+    #[test]
+    fn read_v20_track_uncompressed_override_compensate_scale() {
+        // assist/shovelknight/model/body/c00/model.nuanmb, FingerL11, Transform
+        let buffer = hex!(
+            0000803f 0000803f 0000803f          // scale
+            00000000 00000000 00000000          // translation
+            0000803f bea4c13f_79906ebe f641bebe // rotation
+            01000000                            // compensate scale
+        );
+
+        let data = create_track_data_v20(
+            &TrackV2 {
+                name: "abc".into(),
+                flags: TrackFlags {
+                    track_type: TrackTypeV2::Transform,
+                    compression_type: CompressionType::ConstTransform,
+                },
+                frame_count: 1,
+                transform_flags: AnimTransformFlags::new().with_override_compensate_scale(true),
+                data_offset: 0,
+                data_size: buffer.len() as u64,
+            },
+            &buffer,
+        )
+        .unwrap();
+
+        // TODO: This should test the values, but this overlaps with anim_buffer tests?
+        assert_eq!("abc", data.name);
+        assert_eq!(true, data.scale_options.inherit_scale);
+        // This is disabled by the TransformFlags
+        assert_eq!(false, data.scale_options.compensate_scale);
     }
 
     #[test]
@@ -1195,6 +1514,7 @@ mod tests {
                     inherit_scale: true,
                     compensate_scale: true,
                 },
+                transform_flags: TransformFlags::default(),
             },
         )
         .unwrap();
@@ -1218,12 +1538,13 @@ mod tests {
                     inherit_scale: false,
                     compensate_scale: false,
                 },
+                transform_flags: TransformFlags::default(),
             },
         );
 
         assert!(matches!(
             result,
-            Err(AnimError::UnsupportedTrackScaleOptions {
+            Err(error::Error::UnsupportedTrackScaleOptions {
                 scale_options: ScaleOptions {
                     inherit_scale: false,
                     compensate_scale: false
