@@ -1,5 +1,4 @@
 use binread::{BinRead, BinReaderExt, BinResult};
-use bitbuffer::{BitReadBuffer, BitReadStream};
 use bitvec::prelude::*;
 use itertools::Itertools;
 use std::io::{Cursor, Read, Seek, Write};
@@ -11,10 +10,13 @@ use ssbh_lib::{
     Ptr16, Ptr32, Vector4,
 };
 
-use super::compression::{
-    CompressedBuffer, CompressedHeader, CompressedTrackData, Compression, CompressionFlags,
+use super::{
+    bitutils::{BitReader, BitWriter},
+    compression::{
+        CompressedBuffer, CompressedHeader, CompressedTrackData, Compression, CompressionFlags,
+    },
 };
-use super::{compression::*, AnimError, ScaleOptions, TrackValues, Transform, UvTransform};
+use super::{compression::*, error::Error, ScaleOptions, TrackValues, Transform, UvTransform};
 
 impl TrackValues {
     pub(crate) fn write<W: Write + Seek>(
@@ -23,7 +25,7 @@ impl TrackValues {
         compression: CompressionType,
         inherit_scale: bool,
         compensate_scale: bool,
-    ) -> Result<(), AnimError> {
+    ) -> Result<(), Error> {
         // TODO: Find a way to simplify calculating the default and compression.
         // TODO: Find a way to clean up this code.
         // The default depends on the values.
@@ -72,7 +74,7 @@ impl TrackValues {
                 TrackValues::Transform(values) => {
                     // Uncompressed transform tracks don't support disabling scale inheritance.
                     if !inherit_scale {
-                        return Err(AnimError::UnsupportedTrackScaleOptions {
+                        return Err(Error::UnsupportedTrackScaleOptions {
                             scale_options: ScaleOptions {
                                 inherit_scale,
                                 compensate_scale,
@@ -163,15 +165,16 @@ fn create_compressed_buffer<T: CompressedData>(
     // Construct a single buffer and keep incrementing the bit index.
     // This essentially creates a bit writer buffered with u8 elements.
     // We already know the exact size, so there's no need to reallocate.
-    let mut bits = BitVec::<Lsb0, u8>::new();
+    // TODO: Can we reserve size and append instead?
+    let mut bits = BitVec::<u8, Lsb0>::new();
     bits.resize(values.len() * compression.bit_count(flags) as usize, false);
+    let mut writer = BitWriter::new(bits);
 
-    let mut bit_index = 0;
     for v in values {
-        v.compress(&mut bits, &mut bit_index, compression, flags);
+        v.compress(&mut writer, compression, flags);
     }
 
-    bits.into_vec()
+    writer.into_bytes()
 }
 
 fn read_uncompressed<R: Read + Seek, T: BinRead>(
@@ -190,10 +193,10 @@ pub fn read_track_values(
     track_data: &[u8],
     flags: TrackFlags,
     count: usize,
-) -> Result<(TrackValues, bool, bool), AnimError> {
+) -> Result<(TrackValues, bool, bool), Error> {
     // TODO: Are Const, ConstTransform, and Direct all the same?
     // TODO: Can frame count be higher than 1 for Const and ConstTransform?
-    use crate::anim_data::TrackType as TrackTy;
+    use crate::anim_data::TrackTypeV2 as TrackTy;
     use crate::anim_data::TrackValues as Values;
 
     let mut reader = Cursor::new(track_data);
@@ -286,10 +289,10 @@ pub fn read_track_values(
     Ok((values, inherit_scale, compensate_scale))
 }
 
-fn read_compressed<R: Read + Seek, T: CompressedData + std::fmt::Debug>(
+fn read_compressed<R: Read + Seek, T: CompressedData>(
     reader: &mut R,
     frame_count: usize,
-) -> Result<Vec<T>, AnimError> {
+) -> Result<Vec<T>, Error> {
     let data: CompressedTrackData<T> = reader.read_le()?;
     let values = read_compressed_inner(data, frame_count)?;
     Ok(values)
@@ -298,36 +301,71 @@ fn read_compressed<R: Read + Seek, T: CompressedData + std::fmt::Debug>(
 fn read_compressed_inner<T: CompressedData>(
     data: CompressedTrackData<T>,
     frame_count: usize,
-) -> Result<Vec<T>, AnimError> {
-    // TODO: Return an error if the header has null pointers.
+) -> Result<Vec<T>, Error> {
+    // Check for unexpected compression flags.
+    // This is either an unresearched flag or an improperly compressed file.
+    let expected_bit_count = data.compression.bit_count(data.header.flags) as usize;
+    if data.header.bits_per_entry as usize != expected_bit_count {
+        return Err(Error::UnexpectedBitCount {
+            expected: expected_bit_count,
+            actual: data.header.bits_per_entry as usize,
+        });
+    }
+
+    let buffer = &data
+        .header
+        .compressed_data
+        .as_ref()
+        .ok_or(Error::MalformedCompressionHeader)?
+        .0;
+
     // Decompress values.
-    let bit_buffer = BitReadBuffer::new(
-        &data.header.compressed_data.as_ref().unwrap().0,
-        bitbuffer::LittleEndian,
-    );
-    let mut bit_reader = BitReadStream::new(bit_buffer);
+    let mut reader = BitReader::from_slice(buffer);
+
+    // Encode a repeated value as a single "frame".
+    // TODO: Investigate the side effects of forcing uncompressed on save.
+    // This prevents a potential out of memory or lengthy loop.
+    // This case doesn't occur in any of Smash Ultimate's game files.
+    let actual_count = if expected_bit_count == 0 && frame_count > 0 {
+        1
+    } else {
+        frame_count
+    };
+
     let mut values = Vec::new();
-    for _ in 0..frame_count {
+    for _ in 0..actual_count {
         let value = T::decompress(
-            &mut bit_reader,
+            &mut reader,
             &data.compression,
-            data.header.default_data.as_ref().unwrap(),
+            data.header
+                .default_data
+                .as_ref()
+                .ok_or(Error::MalformedCompressionHeader)?,
             T::get_args(&data.header),
         )?;
+
         values.push(value);
     }
+
     Ok(values)
 }
 
 fn read_compressed_transforms<R: Read + Seek>(
     reader: &mut R,
     frame_count: usize,
-) -> Result<(Vec<UncompressedTransform>, bool, bool), AnimError> {
+) -> Result<(Vec<UncompressedTransform>, bool, bool), Error> {
     let data: CompressedTrackData<UncompressedTransform> = reader.read_le()?;
 
     // TODO: Is this the best way to handle scale settings?
     let inherit_scale = data.header.flags.scale_type() != ScaleType::ScaleNoInheritance;
-    let compensate_scale = data.header.default_data.unwrap().compensate_scale != 0;
+
+    let compensate_scale = data
+        .header
+        .default_data
+        .as_ref()
+        .ok_or(Error::MalformedCompressionHeader)?
+        .compensate_scale
+        != 0;
 
     let values = read_compressed_inner(data, frame_count)?;
 
@@ -339,7 +377,7 @@ mod tests {
     use super::*;
     use crate::{anim_data::Transform, assert_hex_eq};
     use hexlit::hex;
-    use ssbh_lib::{formats::anim::TrackType, Vector3};
+    use ssbh_lib::{formats::anim::TrackTypeV2, Vector3};
 
     #[test]
     fn read_constant_vector4_single_frame() {
@@ -348,7 +386,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::Vector4,
+                track_type: TrackTypeV2::Vector4,
                 compression_type: CompressionType::Constant,
             },
             1,
@@ -388,7 +426,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::UvTransform,
+                track_type: TrackTypeV2::UvTransform,
                 compression_type: CompressionType::Constant,
             },
             1,
@@ -461,7 +499,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::UvTransform,
+                track_type: TrackTypeV2::UvTransform,
                 compression_type: CompressionType::Compressed,
             },
             4,
@@ -535,7 +573,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::UvTransform,
+                track_type: TrackTypeV2::UvTransform,
                 compression_type: CompressionType::Compressed,
             },
             37,
@@ -614,7 +652,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::PatternIndex,
+                track_type: TrackTypeV2::PatternIndex,
                 compression_type: CompressionType::Constant,
             },
             1,
@@ -656,7 +694,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::PatternIndex,
+                track_type: TrackTypeV2::PatternIndex,
                 compression_type: CompressionType::Compressed,
             },
             8,
@@ -674,6 +712,25 @@ mod tests {
         ));
     }
 
+    #[test]
+    fn read_compressed_pattern_index_zero_bit_count() {
+        let data = hex!(
+            00000000 22000000 00000004 00000000      // header
+            00000000 00000000 00000000 00000000      // compression
+            00000004 00000000                        // default value
+            000000000080000010000000000000ffffff     // compressed values
+        );
+        read_track_values(
+            &data,
+            TrackFlags {
+                track_type: TrackTypeV2::PatternIndex,
+                compression_type: CompressionType::Compressed,
+            },
+            8,
+        )
+        .unwrap();
+    }
+
     #[test]
     fn read_constant_float_single_frame() {
         // assist/shovelknight/model/body/c00/model.nuanmb, asf_shovelknight_mat, CustomFloat8
@@ -681,7 +738,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::Float,
+                track_type: TrackTypeV2::Float,
                 compression_type: CompressionType::Constant,
             },
             1,
@@ -710,6 +767,34 @@ mod tests {
         assert_eq!(*writer.get_ref(), hex!(cdcccc3e));
     }
 
+    #[test]
+    fn read_compressed_float_all_equal() {
+        // This is an edge case that doesn't appear in game.
+        // It's possible to have a high frame count with 0 bits per entry.
+        // The default value is used for all entries.
+        // A naive implementation will likely crash.
+        let data = hex!(
+            04000000 20000000 24000000 FFFFFFFF // header
+            cdcccc3e cdcccc3e 10000000 00000000 // compression
+            cdcccc3e                            // default value
+                                                // compressed values
+        );
+        let (values, inherit_scale, compensate_scale) = read_track_values(
+            &data,
+            TrackFlags {
+                track_type: TrackTypeV2::Float,
+                compression_type: CompressionType::Compressed,
+            },
+            0xFFFFFFFF,
+        )
+        .unwrap();
+
+        assert_eq!(false, inherit_scale);
+        assert_eq!(false, compensate_scale);
+
+        assert!(matches!(values, TrackValues::Float(values) if values == vec![0.4]));
+    }
+
     #[test]
     fn read_compressed_float_multiple_frames() {
         // pacman/model/body/c00/model.nuanmb, phong3__phong0__S_CUS_0xa2001001___7__AT_GREATER128___VTC__NORMEXP16___CULLNONE_A_AB_SORT, CustomFloat2
@@ -722,7 +807,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::Float,
+                track_type: TrackTypeV2::Float,
                 compression_type: CompressionType::Compressed,
             },
             5,
@@ -774,7 +859,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::Boolean,
+                track_type: TrackTypeV2::Boolean,
                 compression_type: CompressionType::Constant,
             },
             1,
@@ -810,7 +895,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::Boolean,
+                track_type: TrackTypeV2::Boolean,
                 compression_type: CompressionType::Constant,
             },
             1,
@@ -834,7 +919,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::Boolean,
+                track_type: TrackTypeV2::Boolean,
                 compression_type: CompressionType::Compressed,
             },
             3,
@@ -959,7 +1044,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::Vector4,
+                track_type: TrackTypeV2::Vector4,
                 compression_type: CompressionType::Compressed,
             },
             8,
@@ -1075,7 +1160,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::Transform,
+                track_type: TrackTypeV2::Transform,
                 compression_type: CompressionType::ConstTransform,
             },
             1,
@@ -1178,8 +1263,8 @@ mod tests {
                 z: float_compression,
             },
         };
-        let bit_buffer = BitReadBuffer::new(&data_hex, bitbuffer::LittleEndian);
-        let mut bit_reader = BitReadStream::new(bit_buffer);
+
+        let mut reader = BitReader::from_slice(&data_hex);
 
         let default = UncompressedTransform {
             scale: Vector3::new(4.0, 4.0, 4.0),
@@ -1187,7 +1272,7 @@ mod tests {
             translation: Vector3::new(3.0, 3.0, 3.0),
             compensate_scale: 0,
         };
-        bit_reader
+        reader
             .decompress(&compression, &default, header.flags)
             .unwrap();
     }
@@ -1212,6 +1297,45 @@ mod tests {
         );
     }
 
+    #[test]
+    fn read_compressed_transform_multiple_frames_null_default() {
+        // assist/shovelknight/model/body/c00/model.nuanmb, ArmL, Transform
+        // Default pointer set to 0.
+        let data = hex!(
+            // header
+            04000600 00002b00 cc000000 02000000
+            // scale compression
+            0000803f 0000803f 10000000 00000000
+            0000803f 0000803f 10000000 00000000
+            0000803f 0000803f 10000000 00000000
+            // rotation compression
+            00000000 b9bc433d 0d000000 00000000
+            e27186bd 00000000 0d000000 00000000
+            00000000 ada2273f 10000000 00000000
+            // translation compression
+            16a41d40 16a41d40 10000000 00000000
+            00000000 00000000 10000000 00000000
+            00000000 00000000 10000000 00000000
+            // default value
+            0000803f 0000803f 0000803f
+            00000000 00000000 00000000 0000803f
+            16a41d40 00000000 00000000
+            00000000
+            // compressed values
+            00e0ff03 00f8ff00 e0ff1f
+        );
+
+        let result = read_track_values(
+            &data,
+            TrackFlags {
+                track_type: TrackTypeV2::Transform,
+                compression_type: CompressionType::Compressed,
+            },
+            2,
+        );
+        assert!(matches!(result, Err(Error::MalformedCompressionHeader)));
+    }
+
     #[test]
     fn read_compressed_transform_multiple_frames() {
         // assist/shovelknight/model/body/c00/model.nuanmb, ArmL, Transform
@@ -1242,7 +1366,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::Transform,
+                track_type: TrackTypeV2::Transform,
                 compression_type: CompressionType::Compressed,
             },
             2,
@@ -1411,7 +1535,7 @@ mod tests {
         let (values, inherit_scale, compensate_scale) = read_track_values(
             &data,
             TrackFlags {
-                track_type: TrackType::Transform,
+                track_type: TrackTypeV2::Transform,
                 compression_type: CompressionType::Direct,
             },
             2,
