@@ -1,30 +1,51 @@
 //! Types for working with [Matl] data in .numatb files.
-use std::{
-    convert::{TryFrom, TryInto},
-    io::{Read, Seek, Write},
-};
+//!
+//! # Examples
+//! The parameters for a single material are grouped into a [MatlEntryData].
+//! This examples shows accessing the first rasterizer state for each material.
+//! This is typically [ParamId::RasterizerState0].
+/*!
+```rust no_run
+# fn main() -> Result<(), Box<dyn std::error::Error>> {
+use ssbh_data::prelude::*;
+
+let matl = MatlData::from_file("model.numatb")?;
+
+for entry in matl.entries {
+    println!(
+        "Material: {:?}, Shader: {:?}",
+        entry.material_label, entry.shader_label
+    );
+
+    let rasterizer_state = &entry.rasterizer_states[0];
+    println!("{:?}", rasterizer_state.param_id);
+    println!("{:?}", rasterizer_state.data);
+}
+# Ok(()) }
+```
+ */
+
+use std::convert::TryFrom;
 
 use itertools::Itertools;
 pub use ssbh_lib::formats::matl::{
-    BlendFactor, CullMode, FillMode, FilteringType, MagFilter, MaxAnisotropy, MinFilter, ParamId,
-    WrapMode,
+    BlendFactor, CullMode, FillMode, MagFilter, MaxAnisotropy, MinFilter, ParamId, WrapMode,
 };
 use ssbh_lib::{
     formats::matl::{
-        MatlAttributeV16, MatlBlendStateV16, MatlEntries, MatlEntryV16, MatlRasterizerStateV16,
-        MatlSampler, ParamV16,
+        AttributeV16, BlendStateV16, MatlEntryV16, ParamV16, RasterizerStateV16, Sampler,
     },
-    RelPtr64, SsbhEnum64,
+    RelPtr64, SsbhEnum64, Version,
+};
+use ssbh_lib::{
+    formats::matl::{FilteringType, Matl},
+    SsbhString,
 };
 pub use ssbh_lib::{Color4f, Vector4};
-use ssbh_lib::{Matl, SsbhString};
-use thiserror::Error;
 
 #[cfg(feature = "serde")]
 use serde::{Deserialize, Serialize};
 
-use crate::SsbhData;
-
 pub type BlendStateParam = ParamData<BlendStateData>;
 pub type FloatParam = ParamData<f32>;
 pub type BooleanParam = ParamData<bool>;
@@ -33,37 +54,47 @@ pub type RasterizerStateParam = ParamData<RasterizerStateData>;
 pub type SamplerParam = ParamData<SamplerData>;
 pub type TextureParam = ParamData<String>;
 
-/// Errors while creating a [Matl] from [MatlData].
-#[derive(Error, Debug)]
-pub enum MatlError {
-    /// Creating a [Matl] file for the given version is not supported.
-    #[error(
-        "Creating a version {}.{} matl is not supported.",
-        major_version,
-        minor_version
-    )]
-    UnsupportedVersion {
-        major_version: u16,
-        minor_version: u16,
-    },
+pub mod error {
+    use thiserror::Error;
 
-    /// An error occurred while writing data.
-    #[error(transparent)]
-    Io(#[from] std::io::Error),
+    /// Errors while creating a [Matl](super::Matl) from [MatlData](super::MatlData).
+    #[derive(Debug, Error)]
+    pub enum Error {
+        /// Creating a [Matl](super::Matl) file for the given version is not supported.
+        #[error(
+            "Creating a version {}.{} matl is not supported.",
+            major_version,
+            minor_version
+        )]
+        UnsupportedVersion {
+            major_version: u16,
+            minor_version: u16,
+        },
+
+        /// An error occurred while writing data.
+        #[error(transparent)]
+        Io(#[from] std::io::Error),
+    }
 }
 
 /// The data associated with a [Matl] file.
 /// The supported version is 1.6.
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug, PartialEq)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct MatlData {
     pub major_version: u16,
     pub minor_version: u16,
     pub entries: Vec<MatlEntryData>,
 }
 
+/// Data associated with a [MatlEntryV16].
+///
+/// Parameters are grouped by their type like [vectors](struct.MatlEntryData.html#structfield.vectors)
+/// or [samplers](struct.MatlEntryData.html#structfield.samplers).
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug, PartialEq)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct MatlEntryData {
     pub material_label: String,
     pub shader_label: String,
@@ -77,8 +108,10 @@ pub struct MatlEntryData {
     // TODO: UV Transform?
 }
 
+/// A material value identified by [param_id](struct.ParamData.html#structfield.param_id).
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug, PartialEq)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct ParamData<T> {
     // TODO: Is it worth restricting param id by type?
     // This would prevent creating a Vector4 param with CustomFloat0's ID.
@@ -87,8 +120,10 @@ pub struct ParamData<T> {
 }
 
 // TODO: Derive default for these types to make them easier to use.
+/// Data associated with a [Sampler].
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug, PartialEq)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct SamplerData {
     pub wraps: WrapMode,
     pub wrapt: WrapMode,
@@ -104,10 +139,31 @@ pub struct SamplerData {
     pub max_anisotropy: Option<MaxAnisotropy>,
 }
 
+impl Default for SamplerData {
+    fn default() -> Self {
+        // Standard texture filtering and wrapping.
+        Self {
+            wraps: WrapMode::ClampToEdge,
+            wrapt: WrapMode::ClampToEdge,
+            wrapr: WrapMode::ClampToEdge,
+            min_filter: MinFilter::LinearMipmapLinear,
+            mag_filter: MagFilter::Linear,
+            border_color: Color4f {
+                r: 0.0,
+                g: 0.0,
+                b: 0.0,
+                a: 0.0,
+            },
+            lod_bias: 0.0,
+            max_anisotropy: None,
+        }
+    }
+}
+
 // TODO: Should data loss from unsupported fields be an error?
 // Just select the most common unk values in Smash Ultimate for now.
-impl From<MatlSampler> for SamplerData {
-    fn from(v: MatlSampler) -> Self {
+impl From<Sampler> for SamplerData {
+    fn from(v: Sampler) -> Self {
         Self {
             wraps: v.wraps,
             wrapt: v.wrapt,
@@ -125,13 +181,13 @@ impl From<MatlSampler> for SamplerData {
     }
 }
 
-impl From<SamplerData> for MatlSampler {
+impl From<SamplerData> for Sampler {
     fn from(v: SamplerData) -> Self {
         Self::from(&v)
     }
 }
 
-impl From<&SamplerData> for MatlSampler {
+impl From<&SamplerData> for Sampler {
     fn from(v: &SamplerData) -> Self {
         Self {
             wraps: v.wraps,
@@ -153,24 +209,37 @@ impl From<&SamplerData> for MatlSampler {
     }
 }
 
+/// Data associated with a [BlendStateV16].
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug, PartialEq)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct BlendStateData {
     pub source_color: BlendFactor,
     pub destination_color: BlendFactor,
     pub alpha_sample_to_coverage: bool,
 }
 
+impl Default for BlendStateData {
+    fn default() -> Self {
+        // No alpha blending.
+        Self {
+            source_color: BlendFactor::One,
+            destination_color: BlendFactor::Zero,
+            alpha_sample_to_coverage: false,
+        }
+    }
+}
+
 // TODO: Should data loss from unsupported fields be an error?
 // Just select the most common unk values in Smash Ultimate for now.
-impl From<MatlBlendStateV16> for BlendStateData {
-    fn from(v: MatlBlendStateV16) -> Self {
+impl From<BlendStateV16> for BlendStateData {
+    fn from(v: BlendStateV16) -> Self {
         Self::from(&v)
     }
 }
 
-impl From<&MatlBlendStateV16> for BlendStateData {
-    fn from(v: &MatlBlendStateV16) -> Self {
+impl From<&BlendStateV16> for BlendStateData {
+    fn from(v: &BlendStateV16) -> Self {
         Self {
             source_color: v.source_color,
             destination_color: v.destination_color,
@@ -179,13 +248,13 @@ impl From<&MatlBlendStateV16> for BlendStateData {
     }
 }
 
-impl From<BlendStateData> for MatlBlendStateV16 {
+impl From<BlendStateData> for BlendStateV16 {
     fn from(v: BlendStateData) -> Self {
         Self::from(&v)
     }
 }
 
-impl From<&BlendStateData> for MatlBlendStateV16 {
+impl From<&BlendStateData> for BlendStateV16 {
     fn from(v: &BlendStateData) -> Self {
         Self {
             source_color: v.source_color,
@@ -202,24 +271,37 @@ impl From<&BlendStateData> for MatlBlendStateV16 {
     }
 }
 
+/// Data associated with a [RasterizerStateV16].
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug, PartialEq)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct RasterizerStateData {
     pub fill_mode: FillMode,
     pub cull_mode: CullMode,
     pub depth_bias: f32,
 }
 
+impl Default for RasterizerStateData {
+    fn default() -> Self {
+        // Solid shading.
+        Self {
+            fill_mode: FillMode::Solid,
+            cull_mode: CullMode::Back,
+            depth_bias: 0.0,
+        }
+    }
+}
+
 // TODO: Should data loss from unsupported fields be an error?
 // Just select the most common unk values in Smash Ultimate for now.
-impl From<MatlRasterizerStateV16> for RasterizerStateData {
-    fn from(v: MatlRasterizerStateV16) -> Self {
+impl From<RasterizerStateV16> for RasterizerStateData {
+    fn from(v: RasterizerStateV16) -> Self {
         Self::from(&v)
     }
 }
 
-impl From<&MatlRasterizerStateV16> for RasterizerStateData {
-    fn from(v: &MatlRasterizerStateV16) -> Self {
+impl From<&RasterizerStateV16> for RasterizerStateData {
+    fn from(v: &RasterizerStateV16) -> Self {
         Self {
             fill_mode: v.fill_mode,
             cull_mode: v.cull_mode,
@@ -228,13 +310,13 @@ impl From<&MatlRasterizerStateV16> for RasterizerStateData {
     }
 }
 
-impl From<RasterizerStateData> for MatlRasterizerStateV16 {
+impl From<RasterizerStateData> for RasterizerStateV16 {
     fn from(v: RasterizerStateData) -> Self {
         Self::from(&v)
     }
 }
 
-impl From<&RasterizerStateData> for MatlRasterizerStateV16 {
+impl From<&RasterizerStateData> for RasterizerStateV16 {
     fn from(v: &RasterizerStateData) -> Self {
         Self {
             fill_mode: v.fill_mode,
@@ -247,28 +329,6 @@ impl From<&RasterizerStateData> for MatlRasterizerStateV16 {
     }
 }
 
-impl SsbhData for MatlData {
-    type WriteError = MatlError;
-
-    fn from_file<P: AsRef<std::path::Path>>(path: P) -> Result<Self, Box<dyn std::error::Error>> {
-        Matl::from_file(path)?.try_into().map_err(Into::into)
-    }
-
-    fn read<R: Read + Seek>(reader: &mut R) -> Result<Self, Box<dyn std::error::Error>> {
-        Matl::read(reader)?.try_into().map_err(Into::into)
-    }
-
-    fn write<W: Write + Seek>(&self, writer: &mut W) -> Result<(), Self::WriteError> {
-        Matl::try_from(self)?.write(writer).map_err(Into::into)
-    }
-
-    fn write_to_file<P: AsRef<std::path::Path>>(&self, path: P) -> Result<(), Self::WriteError> {
-        Matl::try_from(self)?
-            .write_to_file(path)
-            .map_err(Into::into)
-    }
-}
-
 // It may be possible to filter a specified enum variant without a macro in the future.
 macro_rules! get_attributes {
     ($attributes:expr, $ty_in:path) => {
@@ -293,7 +353,7 @@ macro_rules! get_attributes {
 }
 
 impl TryFrom<Matl> for MatlData {
-    type Error = MatlError;
+    type Error = error::Error;
 
     fn try_from(value: Matl) -> Result<Self, Self::Error> {
         Self::try_from(&value)
@@ -301,27 +361,26 @@ impl TryFrom<Matl> for MatlData {
 }
 
 impl TryFrom<&Matl> for MatlData {
-    type Error = MatlError;
+    type Error = error::Error;
 
     fn try_from(data: &Matl) -> Result<Self, Self::Error> {
+        let (major_version, minor_version) = data.major_minor_version();
         Ok(Self {
-            major_version: data.major_version,
-            minor_version: data.minor_version,
-            entries: match &data.entries {
-                MatlEntries::EntriesV15(_) => Err(MatlError::UnsupportedVersion {
+            major_version,
+            minor_version,
+            entries: match &data {
+                Matl::V15 { entries: _ } => Err(error::Error::UnsupportedVersion {
                     major_version: 1,
                     minor_version: 5,
                 }),
-                MatlEntries::EntriesV16(entries) => {
-                    Ok(entries.elements.iter().map(Into::into).collect())
-                }
+                Matl::V16 { entries } => Ok(entries.elements.iter().map(Into::into).collect()),
             }?,
         })
     }
 }
 
 impl TryFrom<MatlData> for Matl {
-    type Error = MatlError;
+    type Error = error::Error;
 
     fn try_from(value: MatlData) -> Result<Self, Self::Error> {
         Self::try_from(&value)
@@ -329,17 +388,18 @@ impl TryFrom<MatlData> for Matl {
 }
 
 impl TryFrom<&MatlData> for Matl {
-    type Error = MatlError;
+    type Error = error::Error;
 
     fn try_from(value: &MatlData) -> Result<Self, Self::Error> {
-        // TODO: Errors on unsupported versions?
-        Ok(Self {
-            major_version: value.major_version,
-            minor_version: value.minor_version,
-            entries: MatlEntries::EntriesV16(
-                value.entries.iter().map(Into::into).collect_vec().into(),
-            ),
-        })
+        match (value.major_version, value.minor_version) {
+            (1, 6) => Ok(Self::V16 {
+                entries: value.entries.iter().map(Into::into).collect_vec().into(),
+            }),
+            _ => Err(error::Error::UnsupportedVersion {
+                major_version: value.major_version,
+                minor_version: value.minor_version,
+            }),
+        }
     }
 }
 
@@ -351,25 +411,19 @@ impl From<&MatlEntryV16> for MatlEntryData {
             vectors: get_attributes!(e.attributes.elements, ParamV16::Vector4),
             floats: get_attributes!(e.attributes.elements, ParamV16::Float),
             booleans: get_attributes!(e.attributes.elements, ParamV16::Boolean, |x| x != 0),
-            textures: get_attributes!(
-                e.attributes.elements,
-                ParamV16::MatlString,
-                |x: &SsbhString| x.to_string_lossy()
-            ),
-            samplers: get_attributes!(
-                e.attributes.elements,
-                ParamV16::Sampler,
-                |x: MatlSampler| x.into()
-            ),
+            textures: get_attributes!(e.attributes.elements, ParamV16::String, |x: &SsbhString| x
+                .to_string_lossy()),
+            samplers: get_attributes!(e.attributes.elements, ParamV16::Sampler, |x: Sampler| x
+                .into()),
             blend_states: get_attributes!(
                 e.attributes.elements,
                 ParamV16::BlendState,
-                |x: MatlBlendStateV16| x.into()
+                |x: BlendStateV16| x.into()
             ),
             rasterizer_states: get_attributes!(
                 e.attributes.elements,
                 ParamV16::RasterizerState,
-                |x: MatlRasterizerStateV16| x.into()
+                |x: RasterizerStateV16| x.into()
             ),
         }
     }
@@ -383,31 +437,31 @@ impl From<&MatlEntryData> for MatlEntryV16 {
             attributes: e
                 .blend_states
                 .iter()
-                .map(|a| MatlAttributeV16 {
+                .map(|a| AttributeV16 {
                     param_id: a.param_id,
                     param: a.data.to_param(),
                 })
-                .chain(e.booleans.iter().map(|a| MatlAttributeV16 {
+                .chain(e.booleans.iter().map(|a| AttributeV16 {
                     param_id: a.param_id,
                     param: a.data.to_param(),
                 }))
-                .chain(e.floats.iter().map(|a| MatlAttributeV16 {
+                .chain(e.floats.iter().map(|a| AttributeV16 {
                     param_id: a.param_id,
                     param: a.data.to_param(),
                 }))
-                .chain(e.vectors.iter().map(|a| MatlAttributeV16 {
+                .chain(e.vectors.iter().map(|a| AttributeV16 {
                     param_id: a.param_id,
                     param: a.data.to_param(),
                 }))
-                .chain(e.rasterizer_states.iter().map(|a| MatlAttributeV16 {
+                .chain(e.rasterizer_states.iter().map(|a| AttributeV16 {
                     param_id: a.param_id,
                     param: a.data.to_param(),
                 }))
-                .chain(e.samplers.iter().map(|a| MatlAttributeV16 {
+                .chain(e.samplers.iter().map(|a| AttributeV16 {
                     param_id: a.param_id,
                     param: a.data.to_param(),
                 }))
-                .chain(e.textures.iter().map(|a| MatlAttributeV16 {
+                .chain(e.textures.iter().map(|a| AttributeV16 {
                     param_id: a.param_id,
                     param: a.data.to_param(),
                 }))
@@ -427,7 +481,6 @@ impl ToParam for Vector4 {
     fn to_param(&self) -> SsbhEnum64<ParamV16> {
         SsbhEnum64 {
             data: RelPtr64::new(ParamV16::Vector4(*self)),
-            data_type: 5,
         }
     }
 }
@@ -436,7 +489,6 @@ impl ToParam for f32 {
     fn to_param(&self) -> SsbhEnum64<ParamV16> {
         SsbhEnum64 {
             data: RelPtr64::new(ParamV16::Float(*self)),
-            data_type: 1,
         }
     }
 }
@@ -445,7 +497,6 @@ impl ToParam for bool {
     fn to_param(&self) -> SsbhEnum64<ParamV16> {
         SsbhEnum64 {
             data: RelPtr64::new(ParamV16::Boolean(if *self { 1 } else { 0 })),
-            data_type: 2,
         }
     }
 }
@@ -459,74 +510,68 @@ impl ToParam for String {
 impl ToParam for &str {
     fn to_param(&self) -> SsbhEnum64<ParamV16> {
         SsbhEnum64 {
-            data: RelPtr64::new(ParamV16::MatlString((*self).into())),
-            data_type: 11,
+            data: RelPtr64::new(ParamV16::String((*self).into())),
         }
     }
 }
 
-impl ToParam for MatlSampler {
+impl ToParam for Sampler {
     fn to_param(&self) -> SsbhEnum64<ParamV16> {
         SsbhEnum64 {
             data: RelPtr64::new(ParamV16::Sampler(self.clone())),
-            data_type: 14,
         }
     }
 }
 
 impl ToParam for SamplerData {
     fn to_param(&self) -> SsbhEnum64<ParamV16> {
-        MatlSampler::from(self).to_param()
+        Sampler::from(self).to_param()
     }
 }
 
-impl ToParam for MatlBlendStateV16 {
+impl ToParam for BlendStateV16 {
     fn to_param(&self) -> SsbhEnum64<ParamV16> {
         SsbhEnum64 {
             data: RelPtr64::new(ParamV16::BlendState(self.clone())),
-            data_type: 17,
         }
     }
 }
 
 impl ToParam for BlendStateData {
     fn to_param(&self) -> SsbhEnum64<ParamV16> {
-        MatlBlendStateV16::from(self).to_param()
+        BlendStateV16::from(self).to_param()
     }
 }
 
-impl ToParam for MatlRasterizerStateV16 {
+impl ToParam for RasterizerStateV16 {
     fn to_param(&self) -> SsbhEnum64<ParamV16> {
         SsbhEnum64 {
             data: RelPtr64::new(ParamV16::RasterizerState(self.clone())),
-            data_type: 18,
         }
     }
 }
 
 impl ToParam for RasterizerStateData {
     fn to_param(&self) -> SsbhEnum64<ParamV16> {
-        MatlRasterizerStateV16::from(self).to_param()
+        RasterizerStateV16::from(self).to_param()
     }
 }
 
 #[cfg(test)]
 mod tests {
-    use ssbh_lib::formats::matl::{MatlAttributeV16, MatlEntryV16};
+    use ssbh_lib::formats::matl::{AttributeV16, MatlEntryV16};
 
     use super::*;
 
     #[test]
     fn create_empty_matl_data_1_5() {
-        let result = MatlData::try_from(&Matl {
-            major_version: 1,
-            minor_version: 5,
-            entries: MatlEntries::EntriesV15(Vec::new().into()),
+        let result = MatlData::try_from(Matl::V15 {
+            entries: Vec::new().into(),
         });
 
         assert!(matches!(
             result,
-            Err(MatlError::UnsupportedVersion {
+            Err(error::Error::UnsupportedVersion {
                 major_version: 1,
                 minor_version: 5
             })
@@ -535,10 +580,8 @@ mod tests {
 
     #[test]
     fn create_empty_matl_data_1_6() {
-        let data = MatlData::try_from(&Matl {
-            major_version: 1,
-            minor_version: 6,
-            entries: MatlEntries::EntriesV16(Vec::new().into()),
+        let data = MatlData::try_from(Matl::V16 {
+            entries: Vec::new().into(),
         })
         .unwrap();
 
@@ -549,89 +592,85 @@ mod tests {
 
     #[test]
     fn create_matl_data_single_entry() {
-        let data = MatlData::try_from(&Matl {
-            major_version: 1,
-            minor_version: 6,
-            entries: MatlEntries::EntriesV16(
-                vec![MatlEntryV16 {
-                    material_label: "a".into(),
-                    attributes: vec![
-                        MatlAttributeV16 {
-                            param_id: ParamId::CustomVector13,
-                            param: Vector4::new(1.0, 2.0, 3.0, 4.0).to_param(),
-                        },
-                        MatlAttributeV16 {
-                            param_id: ParamId::CustomFloat5,
-                            param: 0.5.to_param(),
-                        },
-                        MatlAttributeV16 {
-                            param_id: ParamId::CustomBoolean0,
-                            param: true.to_param(),
-                        },
-                        MatlAttributeV16 {
-                            param_id: ParamId::CustomBoolean1,
-                            param: false.to_param(),
-                        },
-                        MatlAttributeV16 {
-                            param_id: ParamId::Texture1,
-                            param: "abc".to_param(),
-                        },
-                        MatlAttributeV16 {
-                            param_id: ParamId::Sampler0,
-                            param: MatlSampler {
-                                wraps: WrapMode::ClampToBorder,
-                                wrapt: WrapMode::ClampToEdge,
-                                wrapr: WrapMode::MirroredRepeat,
-                                min_filter: MinFilter::LinearMipmapLinear,
-                                mag_filter: MagFilter::Nearest,
-                                texture_filtering_type: FilteringType::AnisotropicFiltering,
-                                border_color: Color4f {
-                                    r: 1.0,
-                                    g: 1.0,
-                                    b: 3.0,
-                                    a: 4.0,
-                                },
-                                unk11: 0,
-                                unk12: 0,
-                                lod_bias: -1.0,
-                                max_anisotropy: MaxAnisotropy::Four,
-                            }
-                            .to_param(),
-                        },
-                        MatlAttributeV16 {
-                            param_id: ParamId::BlendState0,
-                            param: MatlBlendStateV16 {
-                                source_color: BlendFactor::DestinationColor,
-                                unk2: 0,
-                                destination_color: BlendFactor::One,
-                                unk4: 0,
-                                unk5: 0,
-                                unk6: 0,
-                                alpha_sample_to_coverage: 1,
-                                unk8: 0,
-                                unk9: 0,
-                                unk10: 0,
-                            }
-                            .to_param(),
-                        },
-                        MatlAttributeV16 {
-                            param_id: ParamId::RasterizerState0,
-                            param: MatlRasterizerStateV16 {
-                                fill_mode: FillMode::Solid,
-                                cull_mode: CullMode::Front,
-                                depth_bias: -5.0,
-                                unk4: 0.0,
-                                unk5: 0.0,
-                                unk6: 0,
-                            }
-                            .to_param(),
-                        },
-                    ]
-                    .into(),
-                    shader_label: "b".into(),
-                }]
+        let data = MatlData::try_from(Matl::V16 {
+            entries: vec![MatlEntryV16 {
+                material_label: "a".into(),
+                attributes: vec![
+                    AttributeV16 {
+                        param_id: ParamId::CustomVector13,
+                        param: Vector4::new(1.0, 2.0, 3.0, 4.0).to_param(),
+                    },
+                    AttributeV16 {
+                        param_id: ParamId::CustomFloat5,
+                        param: 0.5.to_param(),
+                    },
+                    AttributeV16 {
+                        param_id: ParamId::CustomBoolean0,
+                        param: true.to_param(),
+                    },
+                    AttributeV16 {
+                        param_id: ParamId::CustomBoolean1,
+                        param: false.to_param(),
+                    },
+                    AttributeV16 {
+                        param_id: ParamId::Texture1,
+                        param: "abc".to_param(),
+                    },
+                    AttributeV16 {
+                        param_id: ParamId::Sampler0,
+                        param: Sampler {
+                            wraps: WrapMode::ClampToBorder,
+                            wrapt: WrapMode::ClampToEdge,
+                            wrapr: WrapMode::MirroredRepeat,
+                            min_filter: MinFilter::LinearMipmapLinear,
+                            mag_filter: MagFilter::Nearest,
+                            texture_filtering_type: FilteringType::AnisotropicFiltering,
+                            border_color: Color4f {
+                                r: 1.0,
+                                g: 1.0,
+                                b: 3.0,
+                                a: 4.0,
+                            },
+                            unk11: 0,
+                            unk12: 0,
+                            lod_bias: -1.0,
+                            max_anisotropy: MaxAnisotropy::Four,
+                        }
+                        .to_param(),
+                    },
+                    AttributeV16 {
+                        param_id: ParamId::BlendState0,
+                        param: BlendStateV16 {
+                            source_color: BlendFactor::DestinationColor,
+                            unk2: 0,
+                            destination_color: BlendFactor::One,
+                            unk4: 0,
+                            unk5: 0,
+                            unk6: 0,
+                            alpha_sample_to_coverage: 1,
+                            unk8: 0,
+                            unk9: 0,
+                            unk10: 0,
+                        }
+                        .to_param(),
+                    },
+                    AttributeV16 {
+                        param_id: ParamId::RasterizerState0,
+                        param: RasterizerStateV16 {
+                            fill_mode: FillMode::Solid,
+                            cull_mode: CullMode::Front,
+                            depth_bias: -5.0,
+                            unk4: 0.0,
+                            unk5: 0.0,
+                            unk6: 0,
+                        }
+                        .to_param(),
+                    },
+                ]
                 .into(),
-            ),
+                shader_label: "b".into(),
+            }]
+            .into(),
         })
         .unwrap();
 
@@ -708,9 +747,9 @@ mod tests {
         let entry = MatlEntryV16 {
             material_label: "alp_mario_002".into(),
             attributes: vec![
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::BlendState0,
-                    param: MatlBlendStateV16 {
+                    param: BlendStateV16 {
                         source_color: BlendFactor::One,
                         unk2: 0,
                         destination_color: BlendFactor::Zero,
@@ -724,41 +763,41 @@ mod tests {
                     }
                     .to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::CustomBoolean1,
                     param: true.to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::CustomBoolean3,
                     param: true.to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::CustomBoolean4,
                     param: true.to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::CustomFloat8,
                     param: 0.7.to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::CustomVector0,
                     param: Vector4::new(1.0, 0.0, 0.0, 0.0).to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::CustomVector13,
                     param: Vector4::new(1.0, 1.0, 1.0, 1.0).to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::CustomVector14,
                     param: Vector4::new(1.0, 1.0, 1.0, 1.0).to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::CustomVector8,
                     param: Vector4::new(1.0, 1.0, 1.0, 1.0).to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::RasterizerState0,
-                    param: MatlRasterizerStateV16 {
+                    param: RasterizerStateV16 {
                         fill_mode: FillMode::Solid,
                         cull_mode: CullMode::Back,
                         depth_bias: 0.0,
@@ -768,9 +807,9 @@ mod tests {
                     }
                     .to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::Sampler0,
-                    param: MatlSampler {
+                    param: Sampler {
                         wraps: WrapMode::Repeat,
                         wrapt: WrapMode::Repeat,
                         wrapr: WrapMode::Repeat,
@@ -790,9 +829,9 @@ mod tests {
                     }
                     .to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::Sampler4,
-                    param: MatlSampler {
+                    param: Sampler {
                         wraps: WrapMode::Repeat,
                         wrapt: WrapMode::Repeat,
                         wrapr: WrapMode::Repeat,
@@ -812,9 +851,9 @@ mod tests {
                     }
                     .to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::Sampler6,
-                    param: MatlSampler {
+                    param: Sampler {
                         wraps: WrapMode::Repeat,
                         wrapt: WrapMode::Repeat,
                         wrapr: WrapMode::Repeat,
@@ -834,9 +873,9 @@ mod tests {
                     }
                     .to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::Sampler7,
-                    param: MatlSampler {
+                    param: Sampler {
                         wraps: WrapMode::ClampToEdge,
                         wrapt: WrapMode::ClampToEdge,
                         wrapr: WrapMode::ClampToEdge,
@@ -856,19 +895,19 @@ mod tests {
                     }
                     .to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::Texture0,
                     param: "alp_mario_002_col".to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::Texture4,
                     param: "alp_mario_002_nor".to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::Texture6,
                     param: "alp_mario_002_prm".to_param(),
                 },
-                MatlAttributeV16 {
+                AttributeV16 {
                     param_id: ParamId::Texture7,
                     param: "#replace_cubemap".to_param(),
                 },
@@ -1049,7 +1088,6 @@ mod tests {
                 std::mem::discriminant(expected.param.data.as_ref().unwrap()),
                 std::mem::discriminant(actual.param.data.as_ref().unwrap())
             );
-            assert_eq!(expected.param.data_type, actual.param.data_type);
         }
     }
 }
