@@ -1,13 +1,14 @@
 use super::{
     AttributeData, AttributeDataTypeV10Ext, AttributeDataTypeV8Ext, MeshObjectData, VectorData,
 };
-use crate::{get_u8_clamped, write_f16, write_f32, write_u8, write_vector_data};
+
+use super::vector_data::*;
 use half::f16;
 use itertools::Itertools;
 use ssbh_lib::{
     formats::mesh::{
         AttributeDataTypeV10, AttributeDataTypeV8, AttributeUsageV8, AttributeUsageV9,
-        MeshAttributeV10, MeshAttributeV8, MeshAttributeV9, MeshAttributes,
+        AttributeV10, AttributeV8, AttributeV9,
     },
     SsbhArray, SsbhString,
 };
@@ -179,7 +180,6 @@ fn create_attributes_from_data<
     F1: Fn(Vec<(&str, usize, U, V)>, u32) -> Vec<(A, V)>,
     F2: Fn(&A) -> usize + Copy,
     F3: Fn(Vec<V>) -> VersionedVectorData,
-    F4: Fn(SsbhArray<A>) -> MeshAttributes,
 >(
     buffer0_data: Vec<(&str, usize, U, V)>,
     buffer1_data: Vec<(&str, usize, U, V)>,
@@ -187,8 +187,7 @@ fn create_attributes_from_data<
     create_buffer_attributes: F1,
     size_in_bytes: F2,
     versioned_vectors: F3,
-    mesh_attributes: F4,
-) -> ([(u32, VersionedVectorData); 4], MeshAttributes) {
+) -> ([(u32, VersionedVectorData); 4], SsbhArray<A>) {
     // Calculate attribute offsets and buffer data in the appropriate format.
     let buffer0_attributes = create_buffer_attributes(buffer0_data, 0);
     let buffer1_attributes = create_buffer_attributes(buffer1_data, 1);
@@ -209,7 +208,7 @@ fn create_attributes_from_data<
             (stride2, versioned_vectors(Vec::new())),
             (0, versioned_vectors(Vec::new())),
         ],
-        mesh_attributes(attributes0.into()),
+        attributes0.into(),
     )
 }
 
@@ -217,7 +216,7 @@ fn create_attributes_from_data<
 // TODO: Struct for the return type?
 pub fn create_attributes_v8(
     data: &MeshObjectData,
-) -> ([(u32, VersionedVectorData); 4], MeshAttributes) {
+) -> ([(u32, VersionedVectorData); 4], SsbhArray<AttributeV8>) {
     // Create a flattened list of attributes grouped by usage.
     // This ensures the attribute order matches existing conventions.
     let buffer0_data = get_positions_v8(&data.positions, AttributeUsageV8::Position)
@@ -237,15 +236,14 @@ pub fn create_attributes_v8(
         buffer1_data,
         32,
         create_buffer_attributes_v8,
-        |a: &MeshAttributeV8| a.data_type.get_size_in_bytes_v8(),
+        |a: &AttributeV8| a.data_type.get_size_in_bytes_v8(),
         VersionedVectorData::V8,
-        MeshAttributes::AttributesV8,
     )
 }
 
 pub fn create_attributes_v9(
     data: &MeshObjectData,
-) -> ([(u32, VersionedVectorData); 4], MeshAttributes) {
+) -> ([(u32, VersionedVectorData); 4], SsbhArray<AttributeV9>) {
     // Create a flattened list of attributes grouped by usage.
     // This ensures the attribute order matches existing conventions.
     let buffer0_data = get_positions_v9(&data.positions, AttributeUsageV9::Position)
@@ -266,15 +264,15 @@ pub fn create_attributes_v9(
         buffer1_data,
         32,
         create_buffer_attributes_v9,
-        |a: &MeshAttributeV9| a.data_type.get_size_in_bytes_v8(),
+        |a: &AttributeV9| a.data_type.get_size_in_bytes_v8(),
         VersionedVectorData::V8,
-        MeshAttributes::AttributesV9,
     )
 }
 
+// TODO: Fix this.
 pub fn create_attributes_v10(
     data: &MeshObjectData,
-) -> ([(u32, VersionedVectorData); 4], MeshAttributes) {
+) -> ([(u32, VersionedVectorData); 4], SsbhArray<AttributeV10>) {
     // Create a flattened list of attributes grouped by usage.
     // This ensures the attribute order matches existing conventions.
     let buffer0_data = get_positions_v10(&data.positions, AttributeUsageV9::Position)
@@ -295,9 +293,8 @@ pub fn create_attributes_v10(
         buffer1_data,
         0,
         create_buffer_attributes_v10,
-        |a: &MeshAttributeV10| a.data_type.get_size_in_bytes_v10(),
+        |a: &AttributeV10| a.data_type.get_size_in_bytes_v10(),
         VersionedVectorData::V10,
-        MeshAttributes::AttributesV10,
     )
 }
 
@@ -409,39 +406,39 @@ fn create_buffer_attributes<
 fn create_buffer_attributes_v8(
     buffer_data: Vec<(&str, usize, AttributeUsageV8, VectorDataV8)>,
     buffer_index: u32,
-) -> Vec<(MeshAttributeV8, VectorDataV8)> {
+) -> Vec<(AttributeV8, VectorDataV8)> {
     create_buffer_attributes(
         buffer_data,
         buffer_index,
         create_attribute_v8,
         VectorDataV8::data_type,
-        |a: &MeshAttributeV8| a.data_type.get_size_in_bytes_v8(),
+        |a: &AttributeV8| a.data_type.get_size_in_bytes_v8(),
     )
 }
 
 fn create_buffer_attributes_v9(
     buffer_data: Vec<(&str, usize, AttributeUsageV9, VectorDataV8)>,
     buffer_index: u32,
-) -> Vec<(MeshAttributeV9, VectorDataV8)> {
+) -> Vec<(AttributeV9, VectorDataV8)> {
     create_buffer_attributes(
         buffer_data,
         buffer_index,
         create_attribute_v9,
         VectorDataV8::data_type,
-        |a: &MeshAttributeV9| a.data_type.get_size_in_bytes_v8(),
+        |a: &AttributeV9| a.data_type.get_size_in_bytes_v8(),
     )
 }
 
 fn create_buffer_attributes_v10(
     buffer_data: Vec<(&str, usize, AttributeUsageV9, VectorDataV10)>,
     buffer_index: u32,
-) -> Vec<(MeshAttributeV10, VectorDataV10)> {
+) -> Vec<(AttributeV10, VectorDataV10)> {
     create_buffer_attributes(
         buffer_data,
         buffer_index,
         create_attribute_v10,
         VectorDataV10::data_type,
-        |a: &MeshAttributeV10| a.data_type.get_size_in_bytes_v10(),
+        |a: &AttributeV10| a.data_type.get_size_in_bytes_v10(),
     )
 }
 
@@ -452,8 +449,8 @@ fn create_attribute_v8(
     usage: AttributeUsageV8,
     data_type: AttributeDataTypeV8,
     buffer_offset: usize,
-) -> MeshAttributeV8 {
-    MeshAttributeV8 {
+) -> AttributeV8 {
+    AttributeV8 {
         usage,
         data_type,
         buffer_index,
@@ -469,8 +466,8 @@ fn create_attribute_v9(
     usage: AttributeUsageV9,
     data_type: AttributeDataTypeV8,
     buffer_offset: usize,
-) -> MeshAttributeV9 {
-    MeshAttributeV9 {
+) -> AttributeV9 {
+    AttributeV9 {
         usage,
         data_type,
         buffer_index,
@@ -488,8 +485,8 @@ fn create_attribute_v10(
     usage: AttributeUsageV9,
     data_type: AttributeDataTypeV10,
     buffer_offset: usize,
-) -> MeshAttributeV10 {
-    MeshAttributeV10 {
+) -> AttributeV10 {
+    AttributeV10 {
         usage,
         data_type,
         buffer_index,
@@ -720,91 +717,86 @@ mod tests {
         assert_eq!(32, stride2);
         assert_eq!(0, stride3);
 
-        match attributes {
-            MeshAttributes::AttributesV8(a) => {
-                let mut attributes = a.elements.iter();
+        let mut attributes = attributes.elements.iter();
 
-                // Check buffer 0.
-                assert_eq!(
-                    &MeshAttributeV8 {
-                        usage: AttributeUsageV8::Position,
-                        data_type: AttributeDataTypeV8::Float3,
-                        buffer_index: 0,
-                        buffer_offset: 0,
-                        sub_index: 0,
-                    },
-                    attributes.next().unwrap()
-                );
+        // Check buffer 0.
+        assert_eq!(
+            &AttributeV8 {
+                usage: AttributeUsageV8::Position,
+                data_type: AttributeDataTypeV8::Float3,
+                buffer_index: 0,
+                buffer_offset: 0,
+                sub_index: 0,
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV8 {
-                        usage: AttributeUsageV8::Normal,
-                        data_type: AttributeDataTypeV8::Float3,
-                        buffer_index: 0,
-                        buffer_offset: 12,
-                        sub_index: 0,
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV8 {
+                usage: AttributeUsageV8::Normal,
+                data_type: AttributeDataTypeV8::Float3,
+                buffer_index: 0,
+                buffer_offset: 12,
+                sub_index: 0,
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV8 {
-                        usage: AttributeUsageV8::Tangent,
-                        data_type: AttributeDataTypeV8::HalfFloat4,
-                        buffer_index: 0,
-                        buffer_offset: 24,
-                        sub_index: 0,
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV8 {
+                usage: AttributeUsageV8::Tangent,
+                data_type: AttributeDataTypeV8::HalfFloat4,
+                buffer_index: 0,
+                buffer_offset: 24,
+                sub_index: 0,
+            },
+            attributes.next().unwrap()
+        );
 
-                // Check buffer 1.
-                assert_eq!(
-                    &MeshAttributeV8 {
-                        usage: AttributeUsageV8::TextureCoordinate,
-                        data_type: AttributeDataTypeV8::Float2,
-                        buffer_index: 1,
-                        buffer_offset: 0,
-                        sub_index: 0,
-                    },
-                    attributes.next().unwrap()
-                );
+        // Check buffer 1.
+        assert_eq!(
+            &AttributeV8 {
+                usage: AttributeUsageV8::TextureCoordinate,
+                data_type: AttributeDataTypeV8::Float2,
+                buffer_index: 1,
+                buffer_offset: 0,
+                sub_index: 0,
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV8 {
-                        usage: AttributeUsageV8::TextureCoordinate,
-                        data_type: AttributeDataTypeV8::Float2,
-                        buffer_index: 1,
-                        buffer_offset: 8,
-                        sub_index: 1,
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV8 {
+                usage: AttributeUsageV8::TextureCoordinate,
+                data_type: AttributeDataTypeV8::Float2,
+                buffer_index: 1,
+                buffer_offset: 8,
+                sub_index: 1,
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV8 {
-                        usage: AttributeUsageV8::ColorSet,
-                        data_type: AttributeDataTypeV8::Byte4,
-                        buffer_index: 1,
-                        buffer_offset: 16,
-                        sub_index: 0,
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV8 {
+                usage: AttributeUsageV8::ColorSet,
+                data_type: AttributeDataTypeV8::Byte4,
+                buffer_index: 1,
+                buffer_offset: 16,
+                sub_index: 0,
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV8 {
-                        usage: AttributeUsageV8::ColorSet,
-                        data_type: AttributeDataTypeV8::Byte4,
-                        buffer_index: 1,
-                        buffer_offset: 20,
-                        sub_index: 1,
-                    },
-                    attributes.next().unwrap()
-                );
-            }
-            _ => panic!("invalid version"),
-        };
+        assert_eq!(
+            &AttributeV8 {
+                usage: AttributeUsageV8::ColorSet,
+                data_type: AttributeDataTypeV8::Byte4,
+                buffer_index: 1,
+                buffer_offset: 20,
+                sub_index: 1,
+            },
+            attributes.next().unwrap()
+        );
     }
 
     #[test]
@@ -863,133 +855,128 @@ mod tests {
         assert_eq!(32, stride2);
         assert_eq!(0, stride3);
 
-        match attributes {
-            MeshAttributes::AttributesV9(a) => {
-                let mut attributes = a.elements.iter();
-                // Check buffer 0.
-                assert_eq!(
-                    &MeshAttributeV9 {
-                        usage: AttributeUsageV9::Position,
-                        data_type: AttributeDataTypeV8::Float3,
-                        buffer_index: 0,
-                        buffer_offset: 0,
-                        sub_index: 0,
-                        name: "p0".into(),
-                        attribute_names: SsbhArray::new(vec!["p0".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        let mut attributes = attributes.elements.iter();
+        // Check buffer 0.
+        assert_eq!(
+            &AttributeV9 {
+                usage: AttributeUsageV9::Position,
+                data_type: AttributeDataTypeV8::Float3,
+                buffer_index: 0,
+                buffer_offset: 0,
+                sub_index: 0,
+                name: "p0".into(),
+                attribute_names: SsbhArray::new(vec!["p0".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV9 {
-                        usage: AttributeUsageV9::Normal,
-                        data_type: AttributeDataTypeV8::Float3,
-                        buffer_index: 0,
-                        buffer_offset: 12,
-                        sub_index: 0,
-                        name: "n0".into(),
-                        attribute_names: SsbhArray::new(vec!["n0".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV9 {
+                usage: AttributeUsageV9::Normal,
+                data_type: AttributeDataTypeV8::Float3,
+                buffer_index: 0,
+                buffer_offset: 12,
+                sub_index: 0,
+                name: "n0".into(),
+                attribute_names: SsbhArray::new(vec!["n0".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV9 {
-                        usage: AttributeUsageV9::Binormal,
-                        data_type: AttributeDataTypeV8::Float3,
-                        buffer_index: 0,
-                        buffer_offset: 24,
-                        sub_index: 0,
-                        // Using "map1" is a convention likely due to generating binormals from this attribute.
-                        name: "map1".into(),
-                        attribute_names: SsbhArray::new(vec!["b1".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV9 {
+                usage: AttributeUsageV9::Binormal,
+                data_type: AttributeDataTypeV8::Float3,
+                buffer_index: 0,
+                buffer_offset: 24,
+                sub_index: 0,
+                // Using "map1" is a convention likely due to generating binormals from this attribute.
+                name: "map1".into(),
+                attribute_names: SsbhArray::new(vec!["b1".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV9 {
-                        usage: AttributeUsageV9::Binormal,
-                        data_type: AttributeDataTypeV8::Float3,
-                        buffer_index: 0,
-                        buffer_offset: 36,
-                        sub_index: 1,
-                        // Using "uvSet" is a convention likely due to generating binormals from this attribute.
-                        name: "uvSet".into(),
-                        attribute_names: SsbhArray::new(vec!["b2".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV9 {
+                usage: AttributeUsageV9::Binormal,
+                data_type: AttributeDataTypeV8::Float3,
+                buffer_index: 0,
+                buffer_offset: 36,
+                sub_index: 1,
+                // Using "uvSet" is a convention likely due to generating binormals from this attribute.
+                name: "uvSet".into(),
+                attribute_names: SsbhArray::new(vec!["b2".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV9 {
-                        usage: AttributeUsageV9::Tangent,
-                        data_type: AttributeDataTypeV8::HalfFloat4,
-                        buffer_index: 0,
-                        buffer_offset: 48,
-                        sub_index: 0,
-                        // Using "map1" is a convention likely due to generating tangents from this attribute.
-                        name: "map1".into(),
-                        attribute_names: SsbhArray::new(vec!["t0".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV9 {
+                usage: AttributeUsageV9::Tangent,
+                data_type: AttributeDataTypeV8::HalfFloat4,
+                buffer_index: 0,
+                buffer_offset: 48,
+                sub_index: 0,
+                // Using "map1" is a convention likely due to generating tangents from this attribute.
+                name: "map1".into(),
+                attribute_names: SsbhArray::new(vec!["t0".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                // Check buffer 1.
-                assert_eq!(
-                    &MeshAttributeV9 {
-                        usage: AttributeUsageV9::TextureCoordinate,
-                        data_type: AttributeDataTypeV8::Float2,
-                        buffer_index: 1,
-                        buffer_offset: 0,
-                        sub_index: 0,
-                        name: "firstUv".into(),
-                        attribute_names: SsbhArray::new(vec!["firstUv".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        // Check buffer 1.
+        assert_eq!(
+            &AttributeV9 {
+                usage: AttributeUsageV9::TextureCoordinate,
+                data_type: AttributeDataTypeV8::Float2,
+                buffer_index: 1,
+                buffer_offset: 0,
+                sub_index: 0,
+                name: "firstUv".into(),
+                attribute_names: SsbhArray::new(vec!["firstUv".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV9 {
-                        usage: AttributeUsageV9::TextureCoordinate,
-                        data_type: AttributeDataTypeV8::Float2,
-                        buffer_index: 1,
-                        buffer_offset: 8,
-                        sub_index: 1,
-                        name: "secondUv".into(),
-                        attribute_names: SsbhArray::new(vec!["secondUv".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV9 {
+                usage: AttributeUsageV9::TextureCoordinate,
+                data_type: AttributeDataTypeV8::Float2,
+                buffer_index: 1,
+                buffer_offset: 8,
+                sub_index: 1,
+                name: "secondUv".into(),
+                attribute_names: SsbhArray::new(vec!["secondUv".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV9 {
-                        usage: AttributeUsageV9::ColorSet,
-                        data_type: AttributeDataTypeV8::Byte4,
-                        buffer_index: 1,
-                        buffer_offset: 16,
-                        sub_index: 0,
-                        name: "color1".into(),
-                        attribute_names: SsbhArray::new(vec!["color1".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV9 {
+                usage: AttributeUsageV9::ColorSet,
+                data_type: AttributeDataTypeV8::Byte4,
+                buffer_index: 1,
+                buffer_offset: 16,
+                sub_index: 0,
+                name: "color1".into(),
+                attribute_names: SsbhArray::new(vec!["color1".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV9 {
-                        usage: AttributeUsageV9::ColorSet,
-                        data_type: AttributeDataTypeV8::Byte4,
-                        buffer_index: 1,
-                        buffer_offset: 20,
-                        sub_index: 1,
-                        name: "color2".into(),
-                        attribute_names: SsbhArray::new(vec!["color2".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
-            }
-            _ => panic!("invalid version"),
-        };
+        assert_eq!(
+            &AttributeV9 {
+                usage: AttributeUsageV9::ColorSet,
+                data_type: AttributeDataTypeV8::Byte4,
+                buffer_index: 1,
+                buffer_offset: 20,
+                sub_index: 1,
+                name: "color2".into(),
+                attribute_names: SsbhArray::new(vec!["color2".into()]),
+            },
+            attributes.next().unwrap()
+        );
     }
 
     #[test]
@@ -1054,133 +1041,128 @@ mod tests {
         assert_eq!(0, stride2);
         assert_eq!(0, stride3);
 
-        match attributes {
-            MeshAttributes::AttributesV10(a) => {
-                let mut attributes = a.elements.iter();
-                // Check buffer 0.
-                assert_eq!(
-                    &MeshAttributeV10 {
-                        usage: AttributeUsageV9::Position,
-                        data_type: AttributeDataTypeV10::Float3,
-                        buffer_index: 0,
-                        buffer_offset: 0,
-                        sub_index: 0,
-                        name: "p0".into(),
-                        attribute_names: SsbhArray::new(vec!["p0".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        let mut attributes = attributes.elements.iter();
+        // Check buffer 0.
+        assert_eq!(
+            &AttributeV10 {
+                usage: AttributeUsageV9::Position,
+                data_type: AttributeDataTypeV10::Float3,
+                buffer_index: 0,
+                buffer_offset: 0,
+                sub_index: 0,
+                name: "p0".into(),
+                attribute_names: SsbhArray::new(vec!["p0".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV10 {
-                        usage: AttributeUsageV9::Normal,
-                        data_type: AttributeDataTypeV10::Float3,
-                        buffer_index: 0,
-                        buffer_offset: 12,
-                        sub_index: 0,
-                        name: "n0".into(),
-                        attribute_names: SsbhArray::new(vec!["n0".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV10 {
+                usage: AttributeUsageV9::Normal,
+                data_type: AttributeDataTypeV10::Float3,
+                buffer_index: 0,
+                buffer_offset: 12,
+                sub_index: 0,
+                name: "n0".into(),
+                attribute_names: SsbhArray::new(vec!["n0".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV10 {
-                        usage: AttributeUsageV9::Binormal,
-                        data_type: AttributeDataTypeV10::Float3,
-                        buffer_index: 0,
-                        buffer_offset: 24,
-                        sub_index: 0,
-                        // Using "map1" is a convention likely due to generating binormals from this attribute.
-                        name: "map1".into(),
-                        attribute_names: SsbhArray::new(vec!["b1".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV10 {
+                usage: AttributeUsageV9::Binormal,
+                data_type: AttributeDataTypeV10::Float3,
+                buffer_index: 0,
+                buffer_offset: 24,
+                sub_index: 0,
+                // Using "map1" is a convention likely due to generating binormals from this attribute.
+                name: "map1".into(),
+                attribute_names: SsbhArray::new(vec!["b1".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV10 {
-                        usage: AttributeUsageV9::Binormal,
-                        data_type: AttributeDataTypeV10::Float3,
-                        buffer_index: 0,
-                        buffer_offset: 36,
-                        sub_index: 1,
-                        // Using "uvSet" is a convention likely due to generating binormals from this attribute.
-                        name: "uvSet".into(),
-                        attribute_names: SsbhArray::new(vec!["b2".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV10 {
+                usage: AttributeUsageV9::Binormal,
+                data_type: AttributeDataTypeV10::Float3,
+                buffer_index: 0,
+                buffer_offset: 36,
+                sub_index: 1,
+                // Using "uvSet" is a convention likely due to generating binormals from this attribute.
+                name: "uvSet".into(),
+                attribute_names: SsbhArray::new(vec!["b2".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV10 {
-                        usage: AttributeUsageV9::Tangent,
-                        data_type: AttributeDataTypeV10::HalfFloat4,
-                        buffer_index: 0,
-                        buffer_offset: 48,
-                        sub_index: 0,
-                        // Using "map1" is a convention likely due to generating tangents from this attribute.
-                        name: "map1".into(),
-                        attribute_names: SsbhArray::new(vec!["t0".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV10 {
+                usage: AttributeUsageV9::Tangent,
+                data_type: AttributeDataTypeV10::HalfFloat4,
+                buffer_index: 0,
+                buffer_offset: 48,
+                sub_index: 0,
+                // Using "map1" is a convention likely due to generating tangents from this attribute.
+                name: "map1".into(),
+                attribute_names: SsbhArray::new(vec!["t0".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                // Check buffer 1.
-                assert_eq!(
-                    &MeshAttributeV10 {
-                        usage: AttributeUsageV9::TextureCoordinate,
-                        data_type: AttributeDataTypeV10::HalfFloat2,
-                        buffer_index: 1,
-                        buffer_offset: 0,
-                        sub_index: 0,
-                        name: "firstUv".into(),
-                        attribute_names: SsbhArray::new(vec!["firstUv".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        // Check buffer 1.
+        assert_eq!(
+            &AttributeV10 {
+                usage: AttributeUsageV9::TextureCoordinate,
+                data_type: AttributeDataTypeV10::HalfFloat2,
+                buffer_index: 1,
+                buffer_offset: 0,
+                sub_index: 0,
+                name: "firstUv".into(),
+                attribute_names: SsbhArray::new(vec!["firstUv".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV10 {
-                        usage: AttributeUsageV9::TextureCoordinate,
-                        data_type: AttributeDataTypeV10::HalfFloat2,
-                        buffer_index: 1,
-                        buffer_offset: 4,
-                        sub_index: 1,
-                        name: "secondUv".into(),
-                        attribute_names: SsbhArray::new(vec!["secondUv".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV10 {
+                usage: AttributeUsageV9::TextureCoordinate,
+                data_type: AttributeDataTypeV10::HalfFloat2,
+                buffer_index: 1,
+                buffer_offset: 4,
+                sub_index: 1,
+                name: "secondUv".into(),
+                attribute_names: SsbhArray::new(vec!["secondUv".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV10 {
-                        usage: AttributeUsageV9::ColorSet,
-                        data_type: AttributeDataTypeV10::Byte4,
-                        buffer_index: 1,
-                        buffer_offset: 8,
-                        sub_index: 0,
-                        name: "color1".into(),
-                        attribute_names: SsbhArray::new(vec!["color1".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
+        assert_eq!(
+            &AttributeV10 {
+                usage: AttributeUsageV9::ColorSet,
+                data_type: AttributeDataTypeV10::Byte4,
+                buffer_index: 1,
+                buffer_offset: 8,
+                sub_index: 0,
+                name: "color1".into(),
+                attribute_names: SsbhArray::new(vec!["color1".into()]),
+            },
+            attributes.next().unwrap()
+        );
 
-                assert_eq!(
-                    &MeshAttributeV10 {
-                        usage: AttributeUsageV9::ColorSet,
-                        data_type: AttributeDataTypeV10::Byte4,
-                        buffer_index: 1,
-                        buffer_offset: 12,
-                        sub_index: 1,
-                        name: "color2".into(),
-                        attribute_names: SsbhArray::new(vec!["color2".into()]),
-                    },
-                    attributes.next().unwrap()
-                );
-            }
-            _ => panic!("invalid version"),
-        };
+        assert_eq!(
+            &AttributeV10 {
+                usage: AttributeUsageV9::ColorSet,
+                data_type: AttributeDataTypeV10::Byte4,
+                buffer_index: 1,
+                buffer_offset: 12,
+                sub_index: 1,
+                name: "color2".into(),
+                attribute_names: SsbhArray::new(vec!["color2".into()]),
+            },
+            attributes.next().unwrap()
+        );
     }
 
     #[test]
