@@ -1,21 +1,26 @@
 //! Types for working with [MeshEx] data in .numshexb files.
+//!
+//! # File Differences
+//! Unmodified files are not guaranteed to be binary identical after saving.
+//! Some bounding information is recalculated on export and is unlikely to match the original file
+//! due to algorithmic differences and floating point errors.
 use geometry_tools::bounding::{
     calculate_bounding_sphere_from_points, calculate_bounding_sphere_from_spheres,
 };
 use itertools::Itertools;
 use ssbh_lib::formats::meshex::AllData;
+use ssbh_lib::{formats::meshex::MeshEx, Ptr64, Vector3};
 pub use ssbh_lib::{CString, Vector4};
-use ssbh_lib::{MeshEx, Ptr64, Vector3};
 
 #[cfg(feature = "serde")]
 use serde::{Deserialize, Serialize};
 
 use crate::mesh_data::MeshObjectData;
-use crate::SsbhData;
 
 /// The data associated with a [MeshEx] file.
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug, PartialEq)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct MeshExData {
     pub mesh_object_groups: Vec<MeshObjectGroupData>,
 }
@@ -25,7 +30,8 @@ pub struct MeshExData {
 /// For example, if three objects have the same name but unique subindices,
 /// [entry_flags](#structfield.entry_flags) will have 3 elements.
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug, PartialEq)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct MeshObjectGroupData {
     pub bounding_sphere: Vector4,
     pub mesh_object_full_name: String,
@@ -34,7 +40,9 @@ pub struct MeshObjectGroupData {
     pub entry_flags: Vec<EntryFlags>,
 }
 
+/// Rendering related toggles for a [MeshObjectData].
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[derive(Debug, PartialEq, Eq, Clone, Copy)]
 pub struct EntryFlags {
     pub draw_model: bool,
@@ -42,31 +50,6 @@ pub struct EntryFlags {
     // TODO: Preserve remaining flags?
 }
 
-impl SsbhData for MeshExData {
-    type WriteError = std::io::Error;
-
-    fn from_file<P: AsRef<std::path::Path>>(path: P) -> Result<Self, Box<dyn std::error::Error>> {
-        Ok(MeshEx::from_file(path)?.into())
-    }
-
-    fn read<R: std::io::Read + std::io::Seek>(
-        reader: &mut R,
-    ) -> Result<Self, Box<dyn std::error::Error>> {
-        Ok(MeshEx::read(reader)?.into())
-    }
-
-    fn write<W: std::io::Write + std::io::Seek>(
-        &self,
-        writer: &mut W,
-    ) -> Result<(), Self::WriteError> {
-        MeshEx::from(self).write(writer)
-    }
-
-    fn write_to_file<P: AsRef<std::path::Path>>(&self, path: P) -> Result<(), Self::WriteError> {
-        MeshEx::from(self).write_to_file(path)
-    }
-}
-
 impl MeshExData {
     /// Groups `objects` by name and calculates bounding information.
     /// Flags are set to a default value.
@@ -109,6 +92,7 @@ impl MeshExData {
 
 // TODO: Add methods to create MeshExData from points.
 
+// TODO: Is there a cleaner way of writing this?
 // TODO: Should this be public?
 fn strip_mesh_name_tags(full_name: &str) -> String {
     // Strip portions of a mesh object's name that aren't necessary for identification.
@@ -122,8 +106,7 @@ fn strip_mesh_name_tags(full_name: &str) -> String {
             if full_name.ends_with("Shape") {
                 full_name
                     .rfind("Shape")
-                    .map(|end_index| full_name.get(..end_index))
-                    .flatten()
+                    .and_then(|end_index| full_name.get(..end_index))
                     .unwrap_or(full_name)
                     .to_string()
             } else {
@@ -141,10 +124,6 @@ fn strip_mesh_name_tags(full_name: &str) -> String {
     }
 }
 
-// We shouldn't have any errors here?
-// TODO: Do we care about null pointers?
-// TODO: How will we calculate the file length?
-// Just buffer into a vec and use the length?
 impl From<MeshEx> for MeshExData {
     fn from(m: MeshEx) -> Self {
         Self::from(&m)
@@ -152,35 +131,41 @@ impl From<MeshEx> for MeshExData {
 }
 
 impl From<&MeshEx> for MeshExData {
+    // TODO: Should null pointers for non strings be considered an error?
     fn from(m: &MeshEx) -> Self {
         Self {
             mesh_object_groups: m
                 .mesh_object_groups
                 .as_ref()
-                .unwrap()
+                .unwrap_or(&Vec::new())
                 .iter()
                 .enumerate()
                 .map(|(i, g)| MeshObjectGroupData {
                     bounding_sphere: g.bounding_sphere,
+                    // Use empty strings for null pointers.
                     mesh_object_full_name: g
                         .mesh_object_full_name
                         .as_ref()
-                        .unwrap()
-                        .to_string_lossy(),
-                    mesh_object_name: g.mesh_object_name.as_ref().unwrap().to_string_lossy(),
+                        .map(|s| s.to_string_lossy())
+                        .unwrap_or_default(),
+                    mesh_object_name: g
+                        .mesh_object_name
+                        .as_ref()
+                        .map(|s| s.to_string_lossy())
+                        .unwrap_or_default(),
                     entry_flags: m
                         .entries
                         .as_ref()
-                        .unwrap()
+                        .unwrap_or(&Vec::new())
                         .iter()
                         .positions(|e| e.mesh_object_group_index as usize == i)
-                        .map(|entry_index| {
-                            // TODO: Error handling here for invalid indices?
-                            let entry_flags = m.entry_flags.as_ref().unwrap().0[entry_index];
-                            EntryFlags {
+                        .filter_map(|entry_index| {
+                            // TODO: Return an error for invalid indices?
+                            let entry_flags = m.entry_flags.as_ref()?.0.get(entry_index)?;
+                            Some(EntryFlags {
                                 draw_model: entry_flags.draw_model(),
                                 cast_shadow: entry_flags.cast_shadow(),
-                            }
+                            })
                         })
                         .collect(),
                 })
@@ -213,15 +198,6 @@ impl From<&MeshExData> for MeshEx {
                 .collect_vec(),
         );
         Self {
-            // TODO: How do we calculate length without a writer?
-            // Is there some sort of heuristic for calculating it?
-            file_length: 0,
-            entry_count: m
-                .mesh_object_groups
-                .iter()
-                .map(|g| g.entry_flags.len() as u32)
-                .sum(),
-            mesh_object_group_count: m.mesh_object_groups.len() as u32,
             all_data: Ptr64::new(AllData {
                 bounding_sphere: Vector4::new(
                     all_sphere.0.x,
@@ -279,16 +255,13 @@ mod tests {
     use super::*;
 
     use ssbh_lib::{
-        formats::meshex::{AllData, MeshEntry, MeshObjectGroup},
-        MeshEx, Ptr64, Vector3, Vector4,
+        formats::meshex::{AllData, MeshEntry, MeshEx, MeshObjectGroup},
+        Ptr64, Vector3, Vector4,
     };
 
     #[test]
     fn convert_meshex_data() {
         let meshex = MeshEx {
-            file_length: 0,
-            entry_count: 2,
-            mesh_object_group_count: 2,
             all_data: Ptr64::new(AllData {
                 bounding_sphere: Vector4::ZERO,
                 name: Ptr64::new("All".into()),
@@ -365,10 +338,7 @@ mod tests {
         assert_eq!(data, MeshExData::from(&meshex));
 
         let new_meshex = MeshEx::from(&data);
-        // TODO: How to test file length?
         // TODO: Test the all data bounding sphere?
-        assert_eq!(3, new_meshex.entry_count);
-        assert_eq!(2, new_meshex.mesh_object_group_count);
         assert_eq!(
             "All",
             new_meshex
@@ -465,63 +435,32 @@ mod tests {
             MeshObjectData {
                 name: "a_VIS".to_string(),
                 sub_index: 0,
-                parent_bone_name: String::new(),
-                sort_bias: 0,
-                disable_depth_write: false,
-                disable_depth_test: false,
-                vertex_indices: Vec::new(),
                 positions: vec![AttributeData {
                     name: String::new(),
                     data: VectorData::Vector3(vec![[-1.0, -1.0, -1.0]; 3]),
                 }],
-                normals: Vec::new(),
-                binormals: Vec::new(),
-                tangents: Vec::new(),
-                texture_coordinates: Vec::new(),
-                color_sets: Vec::new(),
-                bone_influences: Vec::new(),
+                ..Default::default()
             },
             MeshObjectData {
                 name: "a_VIS".to_string(),
-                sub_index: 0,
-                parent_bone_name: String::new(),
-                sort_bias: 0,
-                disable_depth_write: false,
-                disable_depth_test: false,
-                vertex_indices: Vec::new(),
+                sub_index: 1,
                 positions: vec![AttributeData {
                     name: String::new(),
                     data: VectorData::Vector3(vec![[1.0, 1.0, 1.0]; 3]),
                 }],
-                normals: Vec::new(),
-                binormals: Vec::new(),
-                tangents: Vec::new(),
-                texture_coordinates: Vec::new(),
-                color_sets: Vec::new(),
-                bone_influences: Vec::new(),
+                ..Default::default()
             },
             MeshObjectData {
                 name: "b_VIS".to_string(),
                 sub_index: 0,
-                parent_bone_name: String::new(),
-                sort_bias: 0,
-                disable_depth_write: false,
-                disable_depth_test: false,
-                vertex_indices: Vec::new(),
                 positions: vec![AttributeData {
                     name: String::new(),
                     data: VectorData::Vector3(vec![[0.0, 0.0, 0.0]; 3]),
                 }],
-                normals: Vec::new(),
-                binormals: Vec::new(),
-                tangents: Vec::new(),
-                texture_coordinates: Vec::new(),
-                color_sets: Vec::new(),
-                bone_influences: Vec::new(),
+                ..Default::default()
             },
         ]);
 
-        // TODO: Implement Default for MeshObjectData?
         assert_eq!(2, data.mesh_object_groups.len());
 
         assert_eq!("a", data.mesh_object_groups[0].mesh_object_name);
@@ -531,8 +470,11 @@ mod tests {
                 EntryFlags {
                     draw_model: true,
                     cast_shadow: true
-                };
-                2
+                },
+                EntryFlags {
+                    draw_model: true,
+                    cast_shadow: true
+                },
             ],
             data.mesh_object_groups[0].entry_flags
         );
