@@ -1,37 +1,52 @@
 //! Types for working with [Skel] data in .nusktb files.
+//!
+//! The data for each bone is collected into a [BoneData] struct with a single transformation matrix.
+//! The missing matrices are recalculated when converting to [Skel] based on the heirachy of [BoneData].
+//!
+//! # File Differences
+//! Unmodified files are not guaranteed to be binary identical after saving.
+//! Calculated matrices may differ from the originals due to slightly different algorithms and floating point errors.
+//! These errors are very small in practice but may cause gameplay differences such as online desyncs.
 use std::{
     collections::HashSet,
-    convert::TryInto,
-    io::{Read, Seek},
-    path::Path,
+    convert::{TryFrom, TryInto},
 };
 
 use glam::Mat4;
+pub use ssbh_lib::formats::skel::BillboardType;
 use ssbh_lib::{
-    formats::skel::{BillboardType, Skel, SkelBoneEntry, SkelEntryFlags},
-    Matrix4x4,
+    formats::skel::{Skel, SkelBoneEntry, SkelEntryFlags},
+    Matrix4x4, Version,
 };
+
 use thiserror::Error;
 
 #[cfg(feature = "serde")]
 use serde::{Deserialize, Serialize};
 
-use crate::{create_ssbh_array, SsbhData};
+use crate::create_ssbh_array;
 
 // TODO: Add methods to SkelData to find the index of a given bone?
 
 /// The data associated with a [Skel] file.
 /// The supported version is 1.0.
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct SkelData {
     pub major_version: u16,
     pub minor_version: u16,
     pub bones: Vec<BoneData>,
 }
 
+/// Data associated with a [SkelBoneEntry].
+///
+/// Only the bone's transformation relative to its parent is stored.
+/// The missing transformation matrices are calculated when converting to [Skel]
+/// based on the heirarchy of [BoneData].
 #[cfg_attr(feature = "serde1", derive(Serialize, Deserialize))]
-#[derive(Debug)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(Debug, PartialEq, Clone)]
 pub struct BoneData {
     /// The name of the bone.
     pub name: String,
@@ -40,57 +55,37 @@ pub struct BoneData {
     pub transform: [[f32; 4]; 4],
     /// The index of the parent bone in the bones collection or [None] if this is a root bone with no parents.
     pub parent_index: Option<usize>,
+    // TODO: Make this an Option for clarity?
+    pub billboard_type: BillboardType,
     // TODO: Flags?
 }
 
-impl SsbhData for SkelData {
-    type WriteError = SkelError;
+pub mod error {
+    use super::*;
+    use thiserror::Error;
 
-    fn from_file<P: AsRef<Path>>(path: P) -> Result<Self, Box<dyn std::error::Error>> {
-        let skel = Skel::from_file(path)?;
-        Ok((&skel).into())
+    /// Errors while creating an [Skel] from [SkelData].
+    #[derive(Debug, Error)]
+    pub enum Error {
+        /// Creating a [Skel] file for the given version is not supported.
+        #[error(
+            "Creating a version {}.{} skel is not supported.",
+            major_version,
+            minor_version
+        )]
+        UnsupportedVersion {
+            major_version: u16,
+            minor_version: u16,
+        },
+
+        /// An error occurred while calculating a transformation matrix.
+        #[error(transparent)]
+        BoneTransform(#[from] BoneTransformError),
+
+        /// An error occurred while writing data to a buffer.
+        #[error(transparent)]
+        Io(#[from] std::io::Error),
     }
-
-    fn read<R: Read + Seek>(reader: &mut R) -> Result<Self, Box<dyn std::error::Error>> {
-        let skel = Skel::read(reader)?;
-        Ok((&skel).into())
-    }
-
-    fn write<W: std::io::Write + Seek>(&self, writer: &mut W) -> Result<(), SkelError> {
-        let skel = create_skel(self)?;
-        skel.write(writer)?;
-        Ok(())
-    }
-
-    fn write_to_file<P: AsRef<Path>>(&self, path: P) -> Result<(), SkelError> {
-        let skel = create_skel(self)?;
-        skel.write_to_file(path)?;
-        Ok(())
-    }
-}
-
-/// Errors while creating an [Skel] from [SkelData].
-#[derive(Error, Debug)]
-
-pub enum SkelError {
-    /// Creating a [Skel] file for the given version is not supported.
-    #[error(
-        "Creating a version {}.{} skel is not supported.",
-        major_version,
-        minor_version
-    )]
-    UnsupportedVersion {
-        major_version: u16,
-        minor_version: u16,
-    },
-
-    /// An error occurred while calculating a transformation matrix.
-    #[error(transparent)]
-    BoneTransform(#[from] BoneTransformError),
-
-    /// An error occurred while writing data to a buffer.
-    #[error(transparent)]
-    Io(#[from] std::io::Error),
 }
 
 /// Calculates the transform of `world_transform` relative to `parent_world_transform`.
@@ -149,56 +144,79 @@ fn inv_transform(m: &[[f32; 4]; 4]) -> Matrix4x4 {
     Matrix4x4::from_rows_array(&inv)
 }
 
-fn create_skel(data: &SkelData) -> Result<Skel, SkelError> {
-    let world_transforms = data
-        .bones
-        .iter()
-        .map(|b| data.calculate_world_transform(b))
-        .collect::<Result<Vec<_>, _>>()?;
+impl TryFrom<SkelData> for Skel {
+    type Error = error::Error;
 
-    // TODO: Add a test for this with a few bones.
-    Ok(Skel {
-        major_version: data.major_version,
-        minor_version: data.minor_version,
-        bone_entries: data
+    fn try_from(data: SkelData) -> Result<Self, Self::Error> {
+        Self::try_from(&data)
+    }
+}
+
+impl TryFrom<&SkelData> for Skel {
+    type Error = error::Error;
+
+    fn try_from(data: &SkelData) -> Result<Self, Self::Error> {
+        let world_transforms = data
             .bones
             .iter()
-            .enumerate()
-            .map(|(i, b)| SkelBoneEntry {
-                name: b.name.clone().into(),
-                index: i as u16,
-                parent_index: match b.parent_index {
-                    Some(index) => index as i16,
-                    None => -1,
-                },
-                // TODO: Preserve or calculate flags?
-                flags: SkelEntryFlags {
-                    unk1: 1,
-                    billboard_type: BillboardType::None,
-                },
-            })
-            .collect::<Vec<SkelBoneEntry>>()
-            .into(),
-        world_transforms: create_ssbh_array(&world_transforms, Matrix4x4::from_rows_array),
-        inv_world_transforms: create_ssbh_array(&world_transforms, inv_transform),
-        transforms: create_ssbh_array(&data.bones, |b| Matrix4x4::from_rows_array(&b.transform)),
-        inv_transforms: create_ssbh_array(&data.bones, |b| inv_transform(&b.transform)),
-    })
+            .map(|b| data.calculate_world_transform(b))
+            .collect::<Result<Vec<_>, _>>()?;
+
+        Ok(Skel::V10 {
+            bone_entries: data
+                .bones
+                .iter()
+                .enumerate()
+                .map(|(i, b)| SkelBoneEntry {
+                    name: b.name.clone().into(),
+                    index: i as u16,
+                    parent_index: match b.parent_index {
+                        Some(index) => index as i16,
+                        None => -1,
+                    },
+                    // TODO: Preserve or calculate flags?
+                    flags: SkelEntryFlags {
+                        unk1: 1,
+                        billboard_type: b.billboard_type,
+                    },
+                })
+                .collect::<Vec<SkelBoneEntry>>()
+                .into(),
+            world_transforms: create_ssbh_array(&world_transforms, Matrix4x4::from_rows_array),
+            inv_world_transforms: create_ssbh_array(&world_transforms, inv_transform),
+            transforms: create_ssbh_array(&data.bones, |b| {
+                Matrix4x4::from_rows_array(&b.transform)
+            }),
+            inv_transforms: create_ssbh_array(&data.bones, |b| inv_transform(&b.transform)),
+        })
+    }
+}
+
+impl From<Skel> for SkelData {
+    fn from(skel: Skel) -> Self {
+        Self::from(&skel)
+    }
 }
 
 impl From<&Skel> for SkelData {
     fn from(skel: &Skel) -> Self {
+        let (major_version, minor_version) = skel.major_minor_version();
         Self {
-            major_version: skel.major_version,
-            minor_version: skel.minor_version,
+            major_version,
+            minor_version,
             // TODO: Add additional validation for mismatched array lengths?
-            bones: skel
-                .bone_entries
-                .elements
-                .iter()
-                .zip(skel.transforms.elements.iter())
-                .map(|(b, t)| create_bone_data(b, t))
-                .collect(),
+            bones: match skel {
+                Skel::V10 {
+                    bone_entries,
+                    transforms,
+                    ..
+                } => bone_entries
+                    .elements
+                    .iter()
+                    .zip(transforms.elements.iter())
+                    .map(|(b, t)| create_bone_data(b, t))
+                    .collect(),
+            },
         }
     }
 }
@@ -208,6 +226,7 @@ fn create_bone_data(b: &SkelBoneEntry, transform: &Matrix4x4) -> BoneData {
         name: b.name.to_string_lossy(),
         transform: transform.to_rows_array(),
         parent_index: b.parent_index.try_into().ok(),
+        billboard_type: b.flags.billboard_type,
     }
 }
 
@@ -222,7 +241,7 @@ impl SkelData {
     /// For single bound mesh objects, the object is transformed by the parent bone's world transform.
     /**
     ```rust
-    # use ssbh_data::skel_data::{BoneData, SkelData};
+    # use ssbh_data::skel_data::{BoneData, SkelData, BillboardType};
     # let data = SkelData {
     #     major_version: 1,
     #     minor_version: 0,
@@ -230,6 +249,7 @@ impl SkelData {
     #         name: "Head".to_string(),
     #         transform: [[0f32; 4]; 4],
     #         parent_index: None,
+    #         billboard_type: BillboardType::Disabled,
     #     }],
     # };
     let parent_bone_name = "Head";
@@ -271,7 +291,7 @@ impl SkelData {
 }
 
 /// Errors while calculating [BoneData] transformation matrices.
-#[derive(Error, Debug)]
+#[derive(Debug, Error)]
 pub enum BoneTransformError {
     #[error(
         "Cyclical bone chains are not supported. A cycle was detected at index {}.",
@@ -282,9 +302,101 @@ pub enum BoneTransformError {
 
 #[cfg(test)]
 mod tests {
+    use super::*;
     use approx::relative_eq;
 
-    use super::*;
+    #[test]
+    fn create_skel_no_bones() {
+        let data = SkelData {
+            major_version: 1,
+            minor_version: 0,
+            bones: Vec::new(),
+        };
+
+        let skel = Skel::try_from(data).unwrap();
+        assert!(matches!(
+            skel,
+            Skel::V10 {
+                bone_entries,
+                world_transforms,
+                inv_world_transforms,
+                transforms,
+                inv_transforms
+            }
+            if bone_entries.elements.is_empty()
+                && world_transforms.elements.is_empty()
+                && inv_world_transforms.elements.is_empty()
+                && transforms.elements.is_empty()
+                && inv_transforms.elements.is_empty()
+        ));
+    }
+
+    #[test]
+    fn create_skel_two_bones() {
+        // TODO: Add separate tests for if the matrices are not invertible?
+        let identity = [
+            [1.0, 0.0, 0.0, 0.0],
+            [0.0, 1.0, 0.0, 0.0],
+            [0.0, 0.0, 1.0, 0.0],
+            [0.0, 0.0, 0.0, 1.0],
+        ];
+
+        let data = SkelData {
+            major_version: 1,
+            minor_version: 0,
+            bones: vec![
+                BoneData {
+                    name: "a".to_string(),
+                    transform: identity,
+                    parent_index: None,
+                    billboard_type: BillboardType::Disabled,
+                },
+                BoneData {
+                    name: "b".to_string(),
+                    transform: identity,
+                    parent_index: Some(0),
+                    billboard_type: BillboardType::XAxisViewPointAligned,
+                },
+            ],
+        };
+
+        let skel = Skel::try_from(data).unwrap();
+        assert!(matches!(
+            skel,
+            Skel::V10 {
+                bone_entries,
+                world_transforms,
+                inv_world_transforms,
+                transforms,
+                inv_transforms
+            }
+            if bone_entries.elements
+                == vec![
+                    SkelBoneEntry {
+                        name: "a".into(),
+                        index: 0,
+                        parent_index: -1,
+                        flags: SkelEntryFlags {
+                            unk1: 1,
+                            billboard_type: BillboardType::Disabled
+                        },
+                    },
+                    SkelBoneEntry {
+                        name: "b".into(),
+                        index: 1,
+                        parent_index: 0,
+                        flags: SkelEntryFlags {
+                            unk1: 1,
+                            billboard_type: BillboardType::XAxisViewPointAligned
+                        },
+                    },
+                ]
+                && world_transforms.elements.len() == 2
+                && inv_world_transforms.elements.len() == 2
+                && transforms.elements.len() == 2
+                && inv_transforms.elements.len() == 2
+        ));
+    }
 
     #[test]
     fn create_bone_data_no_parent() {
@@ -294,21 +406,24 @@ mod tests {
             parent_index: -1,
             flags: SkelEntryFlags {
                 unk1: 1,
-                billboard_type: BillboardType::None,
+                billboard_type: BillboardType::XYAxisViewPointAligned,
             },
         };
-        let data = create_bone_data(&b, &Matrix4x4::identity());
-        assert_eq!("abc", data.name);
+
         assert_eq!(
-            [
-                [1.0, 0.0, 0.0, 0.0],
-                [0.0, 1.0, 0.0, 0.0],
-                [0.0, 0.0, 1.0, 0.0],
-                [0.0, 0.0, 0.0, 1.0]
-            ],
-            data.transform
+            BoneData {
+                name: "abc".to_string(),
+                transform: [
+                    [1.0, 0.0, 0.0, 0.0],
+                    [0.0, 1.0, 0.0, 0.0],
+                    [0.0, 0.0, 1.0, 0.0],
+                    [0.0, 0.0, 0.0, 1.0]
+                ],
+                parent_index: None,
+                billboard_type: BillboardType::XYAxisViewPointAligned
+            },
+            create_bone_data(&b, &Matrix4x4::identity())
         );
-        assert_eq!(None, data.parent_index);
     }
 
     #[test]
@@ -321,12 +436,24 @@ mod tests {
             parent_index: -5,
             flags: SkelEntryFlags {
                 unk1: 1,
-                billboard_type: BillboardType::None,
+                billboard_type: BillboardType::Disabled,
             },
         };
-        let data = create_bone_data(&b, &Matrix4x4::identity());
-        assert_eq!("abc", data.name);
-        assert_eq!(None, data.parent_index);
+
+        assert_eq!(
+            BoneData {
+                name: "abc".to_string(),
+                transform: [
+                    [1.0, 0.0, 0.0, 0.0],
+                    [0.0, 1.0, 0.0, 0.0],
+                    [0.0, 0.0, 1.0, 0.0],
+                    [0.0, 0.0, 0.0, 1.0]
+                ],
+                parent_index: None,
+                billboard_type: BillboardType::Disabled
+            },
+            create_bone_data(&b, &Matrix4x4::identity())
+        );
     }
 
     #[test]
@@ -405,6 +532,7 @@ mod tests {
                 name: "root".to_string(),
                 transform,
                 parent_index: None,
+                billboard_type: BillboardType::Disabled,
             }],
         };
 
@@ -423,6 +551,7 @@ mod tests {
                 name: "root".to_string(),
                 transform: [[0.0; 4]; 4],
                 parent_index: Some(0),
+                billboard_type: BillboardType::Disabled,
             }],
         };
 
@@ -444,21 +573,25 @@ mod tests {
                     name: "a".to_string(),
                     transform: [[0.0; 4]; 4],
                     parent_index: None,
+                    billboard_type: BillboardType::Disabled,
                 },
                 BoneData {
                     name: "b".to_string(),
                     transform: [[0.0; 4]; 4],
                     parent_index: Some(2),
+                    billboard_type: BillboardType::Disabled,
                 },
                 BoneData {
                     name: "c".to_string(),
                     transform: [[0.0; 4]; 4],
                     parent_index: Some(1),
+                    billboard_type: BillboardType::Disabled,
                 },
                 BoneData {
                     name: "d".to_string(),
                     transform: [[0.0; 4]; 4],
                     parent_index: Some(2),
+                    billboard_type: BillboardType::Disabled,
                 },
             ],
         };
@@ -487,6 +620,7 @@ mod tests {
                         [0.0, 0.0, 0.0, 1.0],
                     ],
                     parent_index: None,
+                    billboard_type: BillboardType::Disabled,
                 },
                 BoneData {
                     name: "Rot".to_string(),
@@ -497,6 +631,7 @@ mod tests {
                         [0.0, 11.241, 0.268775, 1.0],
                     ],
                     parent_index: Some(0),
+                    billboard_type: BillboardType::Disabled,
                 },
                 BoneData {
                     name: "Hip".to_string(),
@@ -507,6 +642,7 @@ mod tests {
                         [0.0, 0.0, 0.0, 1.0],
                     ],
                     parent_index: Some(1),
+                    billboard_type: BillboardType::Disabled,
                 },
                 BoneData {
                     name: "Waist".to_string(),
@@ -517,6 +653,7 @@ mod tests {
                         [1.38263, 0.0, 0.0, 1.0],
                     ],
                     parent_index: Some(2),
+                    billboard_type: BillboardType::Disabled,
                 },
             ],
         };
